/*-
 * Copyright (c) 1998 Initio Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY INITIO CORPORATION ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL INITIO CORPORATION BE LIABLE FOR ANY DIRECT,INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: a100scsi.c v 1.00 1998/10/08 se Exp $	
 */

#ifdef __NetBSD__
#ifdef _KERNEL
#define KERNEL
#endif
#endif

#include <sys/param.h>

#ifdef KERNEL
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/buf.h>
#include <sys/kernel.h>
#ifdef __NetBSD__
#define bootverbose	1
#endif

#include <vm/vm.h>
#include <vm/pmap.h>
#endif /* KERNEL */

#include <pci/pcivar.h>
#include <pci/pcireg.h>

#include <scsi/scsiconf.h>

#include <machine/clock.h>


#include <pci/a100scsi.h>




#define INT32	  int32
#define U_INT32   u_int32


#define OutB(val, port) 		outb(port, val)
#define OutW(val, port) 		outw(port, val)
#define OutL(val, port) 		outl(port, val)

#define PCI_DEVICE_ID_INICA10 	0x10601101ul
#define PCI_BASEADDR0			0x10


#ifdef KERNEL

#ifdef __NetBSD__
static	int	a100_probe	(struct device *, void *, void *);
static	void	a100_attach	(struct device *, struct device *, void *);
#else /* !__NetBSD */
static	char*	a100_probe( pcici_t tag, pcidi_t type);
static	void	a100_attach( pcici_t tag, int unit);
#endif /* __NetBSD__ */

static int orc_waitChipReady(A100_HCS *hcsp);
static int orc_waitFWReady(A100_HCS *hcsp);
static int orc_waitSCSIRSTdone(A100_HCS *hcsp);
static int orc_waitStopOn(A100_HCS *hcsp);
static int orc_waitHDOoff(A100_HCS *hcsp);
static int orc_waitHDIset(A100_HCS *hcsp,UCHAR *pData);

static int orc_get_FW_version(A100_HCS *hcsp, unsigned short *version);
unsigned long orc_get_echo(A100_HCS *hcsp, unsigned long lValue);
static int orc_set_NVRAM(A100_HCS *hcsp, unsigned char address, unsigned char value);
unsigned char orc_get_NVRAM(A100_HCS *hcsp, unsigned char address,unsigned char *pDataIn);
unsigned char orc_get_Bus_Status(A100_HCS *hcsp);
unsigned char orc_abort_SCB(A100_HCS *hcsp, A100_SCB *pScb);
static int orc_push_SCB(A100_HCS *hcsp, A100_SCB *pScb);
static void orc_send_scb(A100_HCS *hcsp, A100_SCB *scbp);


static int orc_rd_all(A100_HCS *hcsp);
void orc_update_all(A100_HCS *hcsp);
void orc_read_eeprom(A100_HCS *hcsp);
static int orc_load_FW(A100_HCS *hcsp);
static void orc_isr (A100_HCS *pHcs);


static void orc_setup_SCBs(A100_HCS *hcsp);

static int orc_init(A100_HCS *hcsp, A100_SCB *pScb,A100_SG *pSG);

static void orc_append_pend_scb(A100_HCS *pHcs, A100_SCB *scbp);

static A100_SCB *orc_abort_srb(A100_HCS *pHcs, ULONG srbp);

static A100_SCB * orc_alloc_scb(A100_HCS *pHcs);
static void orc_release_scb(A100_HCS *pHcs, A100_SCB *pScb);

static A100_SCB * a100_get_scb(A100_HCS *pHcs, u_int32_t flags);
static void a100_free_scb(A100_HCS *pHcs, A100_SCB *pScb);

static int a100_wait_scb(A100_HCS *pHcs,A100_SCB *pScb,int time);


static void a100_send_scb(A100_HCS *pHcs);
static void a100_exec_scb(A100_HCS *pHcs, A100_SCB *pScb);
static void a100_timeout( void *arg1);
static void a100_done(A100_HCS *pHcs, A100_SCB *pScb);



static	int32_t a100_scsi_cmd( struct scsi_xfer *xs);

static	void	a100_min_phys( struct buf *pbuf);


#endif /* KERNEL */




static u_long	a100_count;

#ifdef __NetBSD__

struct	cfdriver a100_cd = {
	NULL, "ihb", DV_DULL
};

struct cfattach a100_ca = {
	sizeof(A100_HCSstruct ), 
	a100_probe, 
	a100_attach
};

#else /* !__NetBSD__ */
struct	pci_device   a100_device = {
	"ihb",
	a100_probe,
	a100_attach,
	&a100_count,
	NULL
};

DATA_SET (pcidevice_set,a100_device);

#endif /* !__NetBSD__ */


struct scsi_adapter a100_switch = {
	a100_scsi_cmd,
	a100_min_phys,
	0,
	0,
#ifndef __NetBSD__
	0,
	"ihb",
#endif /* !__NetBSD__ */
};

struct scsi_device a100_dev = {
	NULL,			/* Use default error handler */
	NULL,			/* have a queue, served by this */
	NULL,			/* have no async handler */
	NULL,			/* Use default 'done' routine */
#ifndef __NetBSD__
	"ihb",
#endif /* !__NetBSD__ */
};




static NVRAM nvram,*nvramp = &nvram;
static UCHAR dftNvRam[64] = {
	0x01,
	0x11,
	0x60,
	0x10,
	0x00,
	0x01,
	0x11,
	0x60,
	0x10,
	0x00,
	0x00,
	0x01,

	0x01,
	0x01,
	0x00,
	0x00,

	0x07,
	0x83,
	0x20,
	0x0A,
	0x00,
	0x00,

	0xC8,0xC8,0xC8,0xC8,
	0xC8,0xC8,0xC8,0xC8,
	0xC8,0xC8,0xC8,0xC8,
	0xC8,0xC8,0xC8,0xC8,

	0x07,
	0x83,
	0x20,
  	0x0A,
	0x00,
	0x00,

	0xC8,0xC8,0xC8,0xC8,
	0xC8,0xC8,0xC8,0xC8,
	0xC8,0xC8,0xC8,0xC8,
	0xC8,0xC8,0xC8,0xC8,

	0x00,
	0x00,
	0x00,
	0x00
};





static int orc_waitChipReady(A100_HCS *hcsp)
{
	int i;
 	for(i = 0;i<2000;i++)
	{
		if(inb(hcsp->HCS_Base+ORC_HCTRL) & HOSTSTOP)
		return (TRUE);
		DELAY(1000);		/* 1 ms */
	}
	return(FALSE);
}
	

static int orc_waitFWReady(A100_HCS *hcsp)
{
	int i;
 	for(i = 0;i<2000;i++)
	{
		if(inb(hcsp->HCS_Base+ORC_HSTUS) & RREADY)
		return (TRUE);
		DELAY(1000);		/* 1 ms */
	}
	return(FALSE);
}


static int orc_waitSCSIRSTdone(A100_HCS *hcsp)
{
	int i;
 	for(i = 0;i<2000;i++)
	{
		if(!(inb(hcsp->HCS_Base+ORC_HCTRL) & SCSIRST))
		return (TRUE);
		DELAY(1000);		/* 1 ms */
	}
	return(FALSE);
}


static int orc_waitStopOn(A100_HCS *hcsp)
{
	int i;
 	for(i = 0;i<2000;i++)
	{
		if(!(inb(hcsp->HCS_Base+ORC_HCTRL) & SCSIRST))
		return (TRUE);
		DELAY(1000);		/* 1 ms */
	}
	return(FALSE);
}


static int orc_waitHDOoff(A100_HCS *hcsp)
{
	int i;
 	for(i = 0;i<2000;i++)
	{
		if(!(inb(hcsp->HCS_Base+ORC_HCTRL) & HDO))
		return (TRUE);
		DELAY(1000);		/* 1 ms */
	}
	return(FALSE);
}



static int orc_waitHDIset(A100_HCS *hcsp,UCHAR *pData)
{
	int i;
 	for(i = 0;i<2000;i++)
	{
		if((*pData = inb(hcsp->HCS_Base+ORC_HSTUS)) & HDI)
		return (TRUE);
		DELAY(1000);		/* 1 ms */
	}
	return(FALSE);
}



static int orc_get_FW_version(A100_HCS *hcsp, unsigned short *version)
{
	UCHAR *pVersion, bData;

	pVersion = (unsigned char *) version;
	outb(hcsp->HCS_Base+ORC_HDATA, ORC_CMD_VERSION);
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);

	if (orc_waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set	*/
	return (FALSE);
	*pVersion++ = inb(hcsp->HCS_Base+ORC_HDATA);
	outb(hcsp->HCS_Base+ORC_HSTUS, bData);		/* Clear HDI	*/

	if (orc_waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set	*/
	return (FALSE);
	*pVersion = inb(hcsp->HCS_Base+ORC_HDATA);
	outb(hcsp->HCS_Base+ORC_HSTUS, bData);		/* Clear HDI	*/

	return (TRUE);
}


/***************************************************************************/
unsigned long orc_get_echo(A100_HCS *hcsp, unsigned long lValue)
{
	unsigned long lDataIn, *pData;
	UCHAR	bData;

	pData = &lValue;
	outb(hcsp->HCS_Base+ORC_HDATA, ORC_CMD_ECHO);
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);

	outb(hcsp->HCS_Base+ORC_HDATA, *pData++);	/* Write byte 0	*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);

	outb(hcsp->HCS_Base+ORC_HDATA, *pData++);	/* Write byte 1	*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);

	outb(hcsp->HCS_Base+ORC_HDATA, *pData++);	/* Write byte 2	*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);

	outb(hcsp->HCS_Base+ORC_HDATA, *pData++);	/* Write byte 3	*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);

	pData = &lDataIn;

	if (orc_waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set	*/
	return (FALSE);
	*pData++ = inb(hcsp->HCS_Base+ORC_HDATA);
	outb(hcsp->HCS_Base+ORC_HSTUS, bData);		/* Clear HDI	*/


	if (orc_waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set	*/
	return (FALSE);
	*pData++ = inb(hcsp->HCS_Base+ORC_HDATA);
	outb(hcsp->HCS_Base+ORC_HSTUS, bData);		/* Clear HDI	*/


	if (orc_waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set	*/
	return (FALSE);
	*pData++ = inb(hcsp->HCS_Base+ORC_HDATA);
	outb(hcsp->HCS_Base+ORC_HSTUS, bData);		/* Clear HDI	*/


	if (orc_waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set	*/
	return (FALSE);
	*pData++ = inb(hcsp->HCS_Base+ORC_HDATA);
	outb(hcsp->HCS_Base+ORC_HSTUS, bData);		/* Clear HDI	*/

	return (lDataIn);
}

/***************************************************************************/
static int orc_set_NVRAM(A100_HCS *hcsp, unsigned char address, unsigned char value)
{

	outb(hcsp->HCS_Base+ORC_HDATA, ORC_CMD_SET_NVM);/* Write command*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);

	outb(hcsp->HCS_Base+ORC_HDATA, address);	/* Write address*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);

	outb(hcsp->HCS_Base+ORC_HDATA, value);		/* Write value	*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);

	return (TRUE);
}

/***************************************************************************/
unsigned char orc_get_NVRAM(A100_HCS *hcsp, unsigned char address,unsigned char *pDataIn)
{
	unsigned char	 bData;

	outb(hcsp->HCS_Base+ORC_HDATA, ORC_CMD_GET_NVM);/* Write command*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);

	outb(hcsp->HCS_Base+ORC_HDATA, address);	/* Write address*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);


	if (orc_waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set	*/
	return (FALSE);
	*pDataIn = inb(hcsp->HCS_Base+ORC_HDATA);
	outb(hcsp->HCS_Base+ORC_HSTUS, bData);		/* Clear HDI	*/

	return (TRUE);
}


/***************************************************************************/
unsigned char orc_get_Bus_Status(A100_HCS *hcsp)
{
	unsigned char	bStatus, bData;

	outb(hcsp->HCS_Base+ORC_HDATA, ORC_CMD_GET_BUS_STATUS);/* Write command*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)		/* Wait HDO off	*/
	return (FALSE);


	if (orc_waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set	*/
	return (FALSE);
	bStatus = inb(hcsp->HCS_Base+ORC_HDATA);
	outb(hcsp->HCS_Base+ORC_HSTUS, bData);		/* Clear HDI	*/

	return (bStatus);
}


/***************************************************************************/
unsigned char orc_abort_SCB(A100_HCS *hcsp, A100_SCB *pScb)
{
	unsigned char	bStatus, bData;

	outb(hcsp->HCS_Base+ORC_HDATA, ORC_CMD_ABORT_SCB);/* Write command*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)           	  /* Wait HDO off*/
		return (FALSE);

	outb(hcsp->HCS_Base+ORC_HDATA, pScb->SCB_ScbIdx); /* Write address*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == FALSE)	       	  /* Wait HDO off*/
		return (FALSE);


	if (orc_waitHDIset(hcsp, &bData) == FALSE)	  /* Wait HDI set*/
		return (FALSE);
	bStatus = inb(hcsp->HCS_Base+ORC_HDATA);
	outb(hcsp->HCS_Base+ORC_HSTUS, bData);		  /* Clear HDI	*/

	if(bStatus == 0)
		return (TRUE);
	else
		return (FALSE);
}

/***************************************************************************/
static int orc_push_SCB(A100_HCS *hcsp, A100_SCB *pScb)
{
	UCHAR	bData;

	outb(hcsp->HCS_Base+ORC_HDATA, ORC_CMD_ISSUE_SCB);/* Write command*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == 0)			 /* Wait HDO off*/
	return (FALSE);

	outb(hcsp->HCS_Base+ORC_HDATA, pScb->SCB_ScbIdx);/* Write address*/
	outb(hcsp->HCS_Base+ORC_HCTRL, HDO);
	if (orc_waitHDOoff(hcsp) == 0)			 /* Wait HDO off*/
	return (FALSE);


	if (orc_waitHDIset(hcsp, &bData) == 0)		 /* Wait HDI set*/
	return (FALSE);

	outb(hcsp->HCS_Base+ORC_HSTUS, bData);		 /* Clear HDI	*/

	return (TRUE);
}

/***************************************************************************/
static void orc_send_scb(A100_HCS *hcsp, A100_SCB *scbp)
{
	scbp->SCB_Status = SCB_POST;

	outb(hcsp->HCS_Base+ORC_PQUEUE, scbp->SCB_ScbIdx);

}



/***********************************************************************
 Read SCSI H/A configuration parameters from serial EEPROM
************************************************************************/
static int orc_rd_all(A100_HCS *hcsp)
{
	int i;
	UCHAR *np, chksum = 0;

	np = (UCHAR *) nvramp;
	for (i = 0; i < 64; i++,np++) {
		if(orc_get_NVRAM(hcsp, (unsigned char ) i,np) == FALSE)
		return -1;
	}
					/*------ Is ckecksum ok ? ------*/
	np = (UCHAR *) nvramp;
	for (i = 0; i < 63; i++)
	chksum += *np++;
	if (nvramp->CheckSum != (UCHAR) chksum)
	return -1;
    
	return 1;
}

/************************************************************************
 Update SCSI H/A configuration parameters from serial EEPROM
*************************************************************************/
void orc_update_all(A100_HCS *hcsp)		 /* setup default pattern*/
{
	int i;
	UCHAR *np, *np1, chksum = 0;		 /* Calculate checksum first*/
	np = (UCHAR *) dftNvRam;
	for (i = 0; i < 63; i++)
	chksum += *np++;
	*np = chksum;

	np = (UCHAR *) dftNvRam;
	np1 = (UCHAR *) nvramp;
	for (i = 0; i < 64; i++, np++, np1++) {
		if (*np != *np1) {
		orc_set_NVRAM(hcsp, (unsigned char) i, *np);
		}
	}
	return;
}

/*************************************************************************
 Function name  : read_eeprom
**************************************************************************/
void orc_read_eeprom(A100_HCS *hcsp)
{
	if (orc_rd_all(hcsp) != 1) {
		orc_update_all( hcsp);		/* setup default pattern*/
		orc_rd_all(hcsp);		/* load again		*/
	}
}


/***************************************************************************/
static int orc_load_FW(A100_HCS *hcsp)
{
	unsigned long	dData;
	unsigned short	wBIOSAddress, i;
	unsigned char	*pData, bData;


	bData = inb(hcsp->HCS_Base+ORC_GCFG);
	outb(hcsp->HCS_Base+ORC_GCFG, bData | EEPRG);	/* Enable EEPROM programming */
	outb(hcsp->HCS_Base+ORC_EBIOSADR2, 0x00);
	outw(hcsp->HCS_Base+ORC_EBIOSADR0, 0x00);
	if (inb(hcsp->HCS_Base+ORC_EBIOSDATA) != 0x55) {
		outb(hcsp->HCS_Base+ORC_GCFG, bData);	/* Disable EEPROM programming */
		return (FALSE);
	}
	outw(hcsp->HCS_Base+ORC_EBIOSADR0, 0x01);
	if (inb(hcsp->HCS_Base+ORC_EBIOSDATA) != 0xAA) {
		outb(hcsp->HCS_Base+ORC_GCFG, bData);	/* Disable EEPROM programming */
		return (FALSE);
	}

	outb(hcsp->HCS_Base+ORC_RISCCTL, PRGMRST | DOWNLOAD);/* Enable SRAM programming */
	pData = (UCHAR *) &dData;
	dData = 0;					/* Initial FW address to 0 */
	outw(hcsp->HCS_Base+ORC_EBIOSADR0, 0x10);
	*pData = inb(hcsp->HCS_Base+ORC_EBIOSDATA);	/* Read from BIOS */
	outw(hcsp->HCS_Base+ORC_EBIOSADR0, 0x11);
	*(pData + 1) = inb(hcsp->HCS_Base+ORC_EBIOSDATA);/* Read from BIOS */
	outw(hcsp->HCS_Base+ORC_EBIOSADR0, 0x12);
	*(pData + 2) = inb(hcsp->HCS_Base+ORC_EBIOSDATA);/* Read from BIOS */
	outb(hcsp->HCS_Base+ORC_EBIOSADR2, *(pData+2));
	outl(hcsp->HCS_Base+ORC_FWBASEADR, dData);	/* Write FW address */

	wBIOSAddress = (unsigned short) dData;
	for (i = 0, pData = (unsigned char *) &dData;	/* Download the code*/
	 i < 0x1000;			/* Firmware code size = 4K	*/
	 i++, wBIOSAddress++) {
		outw(hcsp->HCS_Base+ORC_EBIOSADR0, wBIOSAddress);
		*pData++ = inb(hcsp->HCS_Base+ORC_EBIOSDATA);/*Read from BIOS */
		if ((i % 4) == 3) {
			outl(hcsp->HCS_Base+ORC_RISCRAM, dData);	/* Write every 4 bytes */
			pData = (unsigned char *) &dData;
		}
	}

	outb(hcsp->HCS_Base+ORC_RISCCTL, PRGMRST | DOWNLOAD);/* Reset program count 0 */
	wBIOSAddress -= 0x1000;		/* Reset the BIOS adddress	*/
	for (i = 0, pData = (unsigned char *) &dData;	/* Check the code	*/
	 i < 0x1000;			/* Firmware code size = 4K	*/
	 i++, wBIOSAddress++) {
		outw(hcsp->HCS_Base+ORC_EBIOSADR0, wBIOSAddress);
		*pData++ = inb(hcsp->HCS_Base+ORC_EBIOSDATA);/* Read from BIOS */
		if ((i % 4) == 3) {
			if (inl(hcsp->HCS_Base+ORC_RISCRAM) != dData) {
				outb(hcsp->HCS_Base+ORC_RISCCTL, PRGMRST);/* Reset program to 0 */
				outb(hcsp->HCS_Base+ORC_GCFG, bData);	/*Disable EEPROM programming*/
				return (FALSE);
			}
			pData = (unsigned char *) &dData;
		}
	}
	outb(hcsp->HCS_Base+ORC_RISCCTL, PRGMRST);	/* Reset program to 0*/
	outb(hcsp->HCS_Base+ORC_GCFG, bData);	/* Disable EEPROM programming */
	return (TRUE);
}


/***************************************************************************/
static void orc_setup_SCBs(A100_HCS *hcsp)
{
	A100_SCB	*pTmpScb, *pPrevScb;
	int		i;

	/* Setup SCB Base and SCB Size registers */
	outb(hcsp->HCS_Base+ORC_SCBSIZE, MAX_SCB );		/* Total number of SCBs */
	outl(hcsp->HCS_Base+ORC_SCBBASE0, hcsp->pPhyScbArray);  /* SCB base address 0	*/
	outl(hcsp->HCS_Base+ORC_SCBBASE1, hcsp->pPhyScbArray);	/* SCB base address 1	*/

	for (i = 0, pTmpScb = hcsp->pScbArray; i < MAX_SCB ; i++, pTmpScb++) {
		pTmpScb->SCB_ScbIdx = i;
	 	pTmpScb->SCB_Reserved0 = 0;
		pTmpScb->SCB_Reserved1 = 0;
		pTmpScb->SCB_SGPAddrHigh = 0;
		pTmpScb->SCB_Link = 0xff;
		pTmpScb->SCB_Status = 0;
		pTmpScb->SCB_TagMsg = 0;
		pTmpScb->SCB_FreeFlag = 0;
		if(i != 0)
			pPrevScb->SCB_NextScb = pTmpScb;
		pPrevScb = pTmpScb;
	}

	hcsp->HCS_FirstAvail = hcsp->pScbArray;
	hcsp->HCS_LastAvail = pPrevScb;


	return;
}

/***************************************************************************/
static int orc_init(A100_HCS *hcsp, A100_SCB *pScb,A100_SG *pSG)
{
	UBYTE	readByte, *readBytep;
	USHORT	revision;
	ULONG	i, j;

	hcsp->pScbArray = pScb;
	hcsp->pPhyScbArray = (ULONG) vtophys(pScb);

	hcsp->SGList = pSG;
	hcsp->pPhySG = (ULONG) vtophys(pSG);

	hcsp->HCS_FirstPend = NULL;
	hcsp->HCS_LastPend = NULL;


	outb(hcsp->HCS_Base+ORC_GIMSK, 0xFF);	/* Disable all interrupt*/
	if (inb(hcsp->HCS_Base+ORC_HSTUS) & RREADY) {
		orc_get_FW_version(hcsp, &revision);
		if (revision == 0xFFFF) {
			outb(hcsp->HCS_Base+ORC_HCTRL, DEVRST);
			if (orc_waitChipReady(hcsp) == FALSE)	/* Wait HOSTSTOP reset	*/
			return (1);
			orc_load_FW(hcsp);			/* Download FW*/
			orc_setup_SCBs(hcsp);			/* Setup SCB Base and SCB Size registers*/

			outb(hcsp->HCS_Base+ORC_HCTRL, 0);	/* clear HOSTSTOP	*/
			if (orc_waitFWReady(hcsp) == FALSE)
			return (1);
		}
		else {
			orc_setup_SCBs(hcsp);			/* Setup SCB Base and SCB Size registers*/
		}
	}
	else {							/* Orchid is not Ready	*/
		outb(hcsp->HCS_Base+ORC_HCTRL, DEVRST);		/* Reset Host Adapter	*/
		if (orc_waitChipReady(hcsp) == FALSE)		/* Wait HOSTSTOP reset	*/
		return (1);
		orc_load_FW(hcsp);				/* Download FW*/
		orc_setup_SCBs(hcsp);				/* Setup SCB Base and SCB Size registers*/
		outb(hcsp->HCS_Base+ORC_HCTRL, HDO);		/* Do Hardware Reset &	*/
		if (orc_waitFWReady(hcsp) == FALSE)
			return (1);

	}

	/*------------- get serial EEProm settting -------*/

	orc_read_eeprom(hcsp);

	if (nvramp->Revision != 1)
	return(-1);

	hcsp->HCS_SCSI_ID = nvramp->SCSI0Id;

	if(inb(hcsp->HCS_Base+ORC_GSTAT) & WIDEBUS)
   		hcsp->HCS_MaxTar = 16;
	else
   		hcsp->HCS_MaxTar = 8;
 
	readBytep = (UCHAR*) &(nvramp->Target00Config);
	for (i = 0; i < 16; readBytep++, i++) {
		hcsp->TargetFlag[i] = (*readBytep) & (~TCF_BUSY);
		hcsp->MaximumTags[i] = MAX_SCB;
		hcsp->ActiveTags[i] = 0;
	}

	if (nvramp->SCSI0Config & NCC_BUSRESET) {		/* Reset SCSI bus*/
		printf("ihb%d:Reset SCSI Bus..\n",hcsp->HCS_Unit);
		outb(hcsp->HCS_Base+ORC_HCTRL, SCSIRST);	/* Do Hardware Reset &	*/
	}
	outb(hcsp->HCS_Base+ORC_GIMSK, 0xFB);			/* enable Reply FIFO interrupt	*/
	return (0);
}






static void orc_append_pend_scb(A100_HCS *pHcs, A100_SCB *scbp)
{
	ULONG       flags;

	scbp->SCB_NextScb = NULL;

	flags = splbio();
	if (pHcs->HCS_LastPend != NULL) {
		pHcs->HCS_LastPend->SCB_NextScb = scbp;
		pHcs->HCS_LastPend = scbp;
	}
	else {
		pHcs->HCS_FirstPend = scbp;
		pHcs->HCS_LastPend = scbp;
	}
	splx(flags);
}





static A100_SCB *orc_abort_srb(A100_HCS *pHcs, ULONG srbp)
{
    
	A100_SCB	*pScb, *pPrevScb;
	ULONG		i;
	int		flags;
    
	flags = splbio();
	pPrevScb = NULL;
	pScb = pHcs->HCS_FirstPend;		/* Check Pend queue */
	while (pScb != NULL) {
		if (pScb->SCB_Srb == srbp) {
	    	if (pPrevScb == NULL) {
			if ((pHcs->HCS_FirstPend = pScb->SCB_NextScb) == NULL)
			pHcs->HCS_LastPend = NULL;
		}
		else {
			pPrevScb->SCB_NextScb = pScb->SCB_NextScb;
			if (pScb == pHcs->HCS_LastPend)
			pHcs->HCS_LastPend = pPrevScb;
		}
		pScb->SCB_NextScb = NULL;
	        a100_free_scb(pHcs, pScb);
		splx(flags);
		return (pScb);
		}

		pPrevScb = pScb;
		pScb = pScb->SCB_NextScb;
	}

	for (i = 0, pScb = pHcs->pScbArray; i < MAX_SCB ; i++, pScb++) {
		if ((pScb->SCB_Status) && (pScb->SCB_Srb == srbp)) {
	        	orc_abort_SCB(pHcs, pScb);
			a100_free_scb(pHcs, pScb);
			splx(flags);
			return (pScb);
		} 
	}
	splx(flags);
	return (NULL);
}






static A100_SCB * orc_alloc_scb(A100_HCS *pHcs)
{
	A100_SCB	*pScb;
	int		opri;

	opri = splbio();

	if ((pScb = pHcs->HCS_FirstAvail) != NULL) {
		if(pScb->SCB_NextScb == NULL) {
		   	pHcs->HCS_FirstAvail = NULL;
		   	pHcs->HCS_LastAvail = NULL;
		}
		else
			pHcs->HCS_FirstAvail = pScb->SCB_NextScb;
		pScb->SCB_NextScb = NULL;
		pScb->SCB_FreeFlag  = 1;
	}
	splx(opri);
	return (pScb);
}




/*
 * An scb (and hence an scb entry on the board) is put onto the
 * free list.
 */
static void orc_release_scb(A100_HCS *pHcs, A100_SCB *pScb)
{       

	int opri;

	if(pScb->SCB_FreeFlag == 0)
		return;
	
	pScb->SCB_FreeFlag = 0;
	pScb->SCB_Srb = NULL;
	pScb->SCB_NextScb = NULL;

	opri = splbio();
	if(pHcs->HCS_LastAvail != NULL) {
		pHcs->HCS_LastAvail->SCB_NextScb = pScb;
		pHcs->HCS_LastAvail = pScb;
        }
	else {
		pHcs->HCS_FirstAvail = pScb;
		pHcs->HCS_LastAvail = pScb;
	}		
	splx(opri);
}


static void orc_isr (A100_HCS *pHcs)
{
	ULONG	bScbIdx;

	if(inb(pHcs->HCS_Base+ORC_RQUEUECNT) == 0)
	return ;

	do{
		bScbIdx = (ULONG)inb(pHcs->HCS_Base+ORC_RQUEUE);
		a100_done(pHcs,&(pHcs->pScbArray[bScbIdx]));
	}while(inb(pHcs->HCS_Base+ORC_RQUEUECNT));

}


/*********************************************************/

/*
 * Get a free scb,   Otherwise either return an error or sleep.
 */
static A100_SCB * a100_get_scb(A100_HCS *pHcs, u_int32_t flags)
{
	A100_SCB *pScb;
	int opri;

	opri = splbio();
	/*
	 * If we can and have to, sleep waiting for one to come free
	 * but only if we can't allocate a new one.
	 */
	while (1) {
		if ((pScb = orc_alloc_scb(pHcs)) == NULL) {
		    if ((flags & SCSI_NOSLEEP) == 0) {
			tsleep((caddr_t)&pHcs->pScbArray, PRIBIO,
				"ihbscb", 0);
			continue;
		    }
		}
		    break;
	}

	splx(opri);

	return (pScb);
}


static void a100_free_scb(A100_HCS *pHcs, A100_SCB *pScb)
{
	orc_release_scb(pHcs, pScb);
	/*
	 * Wake anybody waiting for one to come free.
	 */

	wakeup((caddr_t)&pHcs->pScbArray);
}


/*
 * We have a scb which has been processed by the
 * adaptor, now we look to see how the operation
 * went.
 */
static void
a100_done(A100_HCS *pHcs, A100_SCB *pScb)
{
	struct scsi_xfer *xs = (struct scsi_xfer *)pScb->SCB_Srb;
	
	SC_DEBUG(xs->sc_link, SDEV_DB2, ("a100_done\n"));

	pScb->SCB_Status = 0;
	pScb->SCB_Srb = NULL;

	if(pScb->SCB_TagMsg) {
		if(pHcs->ActiveTags[pScb->SCB_Target] <= 1)
			pHcs->ActiveTags[pScb->SCB_Target] = 0;	
		else
			pHcs->ActiveTags[pScb->SCB_Target]--;

		if(pHcs->MaximumTags[pScb->SCB_Target] > pHcs->ActiveTags[pScb->SCB_Target])
			pHcs->TargetFlag[pScb->SCB_Target] &= ~TCF_BUSY;
	}
	else
		pHcs->TargetFlag[pScb->SCB_Target] &= ~TCF_BUSY;

	if(xs == 0) {
		printf("free scb while xs=0 in a100_done \n");
		a100_free_scb(pHcs, pScb);
		return;
	}

	xs->status = pScb->SCB_TaStat;
	/*
	**	Check the status.
	*/
	if (   (pScb->SCB_HaStat == HOST_OK)
		&& (pScb->SCB_TaStat == TARGET_GOOD)) {
#if 1
		/*
		 * This functionality will be provided by the generic SCSI layer
		 * in FreeBSD 3.0.
		 */
		if (xs->cmd->opcode == INQUIRY) {
			struct scsi_inquiry_data *inq_data;
			/*
			 * Sneak a look at the results of the SCSI Inquiry
			 * command and see if we can do Tagged queing.  This
			 * should really be done by the higher level drivers.
			 */
			inq_data = (struct scsi_inquiry_data *)xs->data;
			if ((inq_data->flags & SID_CmdQue)) {
				if ((pHcs->TargetFlag[pScb->SCB_Target] & TCF_EN_TAG) == 0) {
					printf("ihb%d: target %d Tagged Queuing Device\n",
					pHcs->HCS_Unit, xs->sc_link->target);
					pHcs->TargetFlag[pScb->SCB_Target] |= TCF_EN_TAG;
					xs->sc_link->opennings += 7;
				}
			}
		}
#endif /* A100_TAGENABLE */
#ifndef __NetBSD__
	} else if (xs->flags & SCSI_ERR_OK) {

		/*
		**   Not correct, but errors expected.
		*/
		xs->resid = 0;
#endif /* !__NetBSD__ */
	} else if ((pScb->SCB_HaStat== HOST_OK)
		&& (pScb->SCB_TaStat== TARGET_CHKCOND  )) {
		
		/*
		**   Check condition code
		*/
		xs->error = XS_SENSE;

	} else if ((pScb->SCB_HaStat== HOST_OK)
		&& (pScb->SCB_TaStat== TARGET_BUSY)) {

		/*
		**   Target is busy.
		*/
		xs->error = XS_BUSY;

	} else if ((pScb->SCB_HaStat== HOST_OK)
		&& (pScb->SCB_TaStat== TARGET_QFULL)) {

		if((pHcs->MaximumTags[pScb->SCB_Target] = pHcs->ActiveTags[pScb->SCB_Target]) < 1)
			pHcs->MaximumTags[pScb->SCB_Target] =  1;
		/*
		**   Target is queue full.
		*/
		xs->error = XS_BUSY;
		if (xs->sc_link->opennings > 1) 
			xs->sc_link->opennings--; 

	} else if (pScb->SCB_HaStat== HOST_SEL_TOUT   ) {

		/*
		**   Device failed selection
		*/
		xs->error = XS_SELTIMEOUT;

	} else { /*  Other protocol messes */
		xs->error = XS_TIMEOUT;
	}

	xs->flags |= ITSDONE;

	/*
	 * If we're polling, we rely on the ITS_DONE flag in the xs structure
	 * to know that the command has completed.  Unfortunately, scsi_done
	 * can cause the same xs to get requeued putting us in an infinite
	 * loop.  So, we defer the scsi_done call until the poll routine exits
	 * its loop.  I hate the way this works.
	 */
	if ((xs->flags & SCSI_NOMASK) == 0) {
		untimeout(a100_timeout, (caddr_t) pScb);
		scsi_done(xs);
	}
	a100_free_scb(pHcs, pScb);

	/* check pending queue */
	a100_send_scb(pHcs);
}

static void
a100_timeout(void *arg1)
{

	A100_SCB *pScb = (A100_SCB*) arg1;
	struct scsi_xfer *xs = (struct scsi_xfer *)pScb->SCB_Srb;
	A100_HCS *pHcs; 
	int     unit;
	int s;

	if (xs == NULL) 
		return;

	pHcs = (A100_HCS *)xs->sc_link->adapter_softc;

	s = splbio();
	if (orc_abort_srb(pHcs, (ULONG) xs)) {
		xs->error = XS_TIMEOUT;
		printf("ihb: timed out\n");
		xs->flags |= ITSDONE;
		scsi_done(xs);
	}
	splx(s);
}


static int a100_wait_scb(A100_HCS *pHcs,A100_SCB *pScb,int time)
{
	while(time >0) {
		if(pScb->SCB_Status == 0) {
	 		a100_done(pHcs,pScb);
			return (TRUE);
		}		
		DELAY(1000);	/* 1 ms */
		time--;
	}

	a100_free_scb(pHcs, pScb);
	return (FALSE);

}



static void a100_send_scb(A100_HCS *pHcs)
{
	A100_SCB *pScb, *pPrev;

	pScb = pHcs->HCS_FirstPend;
	for (; pScb != NULL; pPrev = pScb, pScb = pScb->SCB_NextScb) {

		if (pHcs->TargetFlag[pScb->SCB_Target] & TCF_BUSY)
		    continue;

		if (pScb->SCB_TagMsg) {
			pHcs->ActiveTags[pScb->SCB_Target]++;
			if(pHcs->MaximumTags[pScb->SCB_Target] <= pHcs->ActiveTags[pScb->SCB_Target])
				pHcs->TargetFlag[pScb->SCB_Target] |= TCF_BUSY;
		}
		else 
			pHcs->TargetFlag[pScb->SCB_Target] |= TCF_BUSY;

		/* unlink pScb */
		if ( pScb == pHcs->HCS_FirstPend) {
			if ((pHcs->HCS_FirstPend = pScb->SCB_NextScb) == NULL)
				pHcs->HCS_LastPend = NULL;
		}
		else if ((pPrev->SCB_NextScb = pScb->SCB_NextScb) == NULL) {
			pHcs->HCS_LastPend = pPrev;
		}

		orc_send_scb(pHcs, pScb);
	}
	
}


static void a100_exec_scb(A100_HCS *pHcs, A100_SCB *pScb)
{
	ULONG	bScbIdx;
	
	pScb->SCB_TagMsg = 0;
	if(pScb->SCB_CDB[0] == 0x3) {

		/* disable SCSI DISCONNECT if it is "REQUEST SENSE" command */
     		pScb->SCB_Ident &= 0xBF;

	}
	else if((pScb->SCB_Flags & SCF_POLL) == 0) {
		if (pHcs->TargetFlag[pScb->SCB_Target]  & TCF_EN_TAG) {
			pScb->SCB_TagMsg = MSG_STAG;
		}
	}

	/* alway append into pending queue */
	orc_append_pend_scb(pHcs, pScb);

	/* check pending queue */
	a100_send_scb(pHcs);
}


/***********************************************************************
 * Function : static int32 a100_scsi_cmd (struct scsi_xfer *cmd)
 * Purpose : enqueues a SCSI command
 ***********************************************************************/

static int32_t a100_scsi_cmd (struct scsi_xfer * xs)
{
	A100_SCB  *pScb;
	A100_HCS  *pHcs;
	int	  flags;

	pHcs = (A100_HCS *)xs->sc_link->adapter_softc;
	SC_DEBUG(xs->sc_link, SDEV_DB2, ("ihb_scsi_cmd\n"));
	flags = xs->flags;

	/*
	 * get an scb to use. If the transfer
	 * is from a buf (possibly from interrupt time)
	 * then we can't allow it to sleep
	 */
	if ((pScb = a100_get_scb(pHcs, flags)) == NULL) {
		printf("Can't get_scb in cmd \n");
		xs->error = XS_DRIVER_STUFFUP;
		return (TRY_AGAIN_LATER);
	}

	SC_DEBUG(xs->sc_link, SDEV_DB3, ("start scb(%p)\n", pScb));

	pScb->SCB_Target = xs->sc_link->target ;
	pScb->SCB_Lun	= xs->sc_link->lun ;
        
	if(xs->flags & SCSI_DATA_IN)
		pScb->SCB_Flags = SCF_DIN;
	else {
		if(xs->flags & SCSI_DATA_OUT)
			pScb->SCB_Flags = SCF_DOUT;
		else	
			pScb->SCB_Flags = SCF_NO_DCHK;
	}

	
	pScb->SCB_HaStat = 0;
	pScb->SCB_TaStat = 0;
	pScb->SCB_Status = 0;
	pScb->SCB_Link   = 0xFF;

	pScb->SCB_Reserved0 = 0;
	pScb->SCB_Reserved1 = 0;

        pScb->SCB_SenseLen = sizeof(xs->sense);
	pScb->SCB_SensePAddr = vtophys(&xs->sense);
	pScb->SCB_Srb =(ULONG)xs;

	if (flags & SCSI_RESET) {
		pScb->SCB_Opcode  = BusDevRst;
	} 
	else {
		pScb->SCB_Opcode  = ExecSCSI;

		pScb->SCB_CDBLen = xs->cmdlen;
		if(pScb->SCB_CDBLen > 14)
			pScb->SCB_CDBLen = 14;	
		bcopy((UCHAR *) xs->cmd, pScb->SCB_CDB, pScb->SCB_CDBLen);

		pScb->SCB_Ident     = pScb->SCB_Lun | DISC_ALLOW;

		xs->resid = 0;
		xs->status = 0;

		if (pScb->SCB_XferLen = xs->datalen) {
			int		seg;
			u_int32_t	datalen;
			vm_offset_t	vaddr;
			u_int32_t	paddr;
			u_int32_t	nextpaddr;
			A100_SG         *sg;
	
			seg = 0;
			datalen = xs->datalen;
			vaddr = (vm_offset_t)xs->data;
			paddr = vtophys(vaddr);
			sg =&pHcs->SGList[pScb->SCB_ScbIdx << 5];
			pScb->SCB_SGPAddr = (ULONG) vtophys(sg); 

			while ((datalen > 0) && (seg < A100_SG_ENTRY))  {
				/* put in the base address and length */
				sg->SG_Ptr = paddr;
				sg->SG_Len= 0;

				/* do it at least once */
				nextpaddr = paddr;

			while ((datalen > 0) && (paddr == nextpaddr)) {
				u_int32_t	size;
				/*
				 * This page is contiguous (physically)
				 * with the the last, just extend the
				 * length
				 */
				/* how far to the end of the page */
				nextpaddr = (paddr & (~PAGE_MASK)) + PAGE_SIZE;

				/*
				 * Compute the maximum size
				 */
				size = nextpaddr - paddr;
				if (size > datalen)
				        size = datalen;


				sg->SG_Len += size;
				vaddr   += size;
				datalen -= size;
				if (datalen > 0)
					paddr = vtophys(vaddr);
			}
			/*
			 * next page isn't contiguous, finish the seg
			 */
			seg++;
			sg++;
			}

			if (datalen) { 
				/* there's still data, must have run out of segs! */
				printf("ihb_scsi_cmd: more than supposed DMA segs exist \n");
				xs->error = XS_DRIVER_STUFFUP;
				a100_free_scb(pHcs, pScb);
				return (COMPLETE);
			}

			pScb->SCB_SGLen = seg << 3;
#ifdef A100_BROKEN_CACHE
			if (a100_broken_cache)
				INVALIDATE_CACHE();
#endif

		}

	}


	if ((flags & SCSI_NOMASK) == 0) {
		pScb->SCB_Flags &= ~SCF_POLL;    /* disable polling*/
		timeout(a100_timeout, (caddr_t)pScb, (xs->timeout * hz)/1000);
		a100_exec_scb(pHcs, pScb);

		return (SUCCESSFULLY_QUEUED);
        }
        else {
		pScb->SCB_Flags |= SCF_POLL;    /* polling*/
		a100_exec_scb(pHcs, pScb);
		if(a100_wait_scb(pHcs,pScb,30000) == FALSE) 
			return (HAD_ERROR);
		else
       	        	return (COMPLETE);
	}

}


void
a100_min_phys( struct buf *bp )
{
	if (bp->b_bcount > ((A100_SG_ENTRY - 1) * PAGELEN))
		bp->b_bcount = ((A100_SG_ENTRY - 1) * PAGELEN);

#if defined(__NetBSD__)
	minphys(bp);
#endif
}




#ifdef __NetBSD__
int
a100print(aux, name)
	void *aux;
	char *name;
{
	if (name != NULL)
		printf("%s: scsibus ", name);
	return UNCONF;
}

#endif /* __NetBSD__ */



#ifdef __NetBSD__

void
a100_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{

	A100_SCB    *pScb;
	struct pci_attach_args *pa = aux;
	A100_HCS *pHcs = (void *)self;
	bus_io_addr_t iobase;
	bus_io_size_t iosize;
        u_int32_t id;
	pci_intr_handle_t ih;
	const char *intrstr;

	/*
	**	Try to map the controller chip to
	**	virtual and physical memory.
	*/
	if (bus_io_map(pa->pa_bc, iobase, iosize, &ioh))
	    if (pci_io_find(pa->pa_pc, pa->pa_tag, PCI_BASEADR0, &iobase,
				&iosize))
		return;


	if (pci_intr_map(pa->pa_pc, pa->pa_intrtag, pa->pa_intrpin,
			 pa->pa_intrline, &ih)) {
		printf("ihb: couldn't map interrupt\n");
		free(pHcs, M_DEVBUF);
		return;
	}
	intrstr = pci_intr_string(pa->pa_pc, ih);

#ifdef __OpenBSD__
	pHcs->sc_ih = pci_intr_establish(pa->pa_pc, ih, IPL_BIO, orc_isr, pHcs,
					pHcs->sc_dev.dv_xname);
#else
	pHcs->sc_ih = pci_intr_establish(pa->pa_pc, ih, IPL_BIO, orc_isr, pHcs);
#endif

	if (pHcs->sc_ih == NULL) {
		printf("ihb: couldn't establish interrupt")
		if (intrstr != NULL)
			printf(" at %s", intrstr);
		printf("\n");
		free(pHcs, M_DEVBUF);
		return;
	}

	pHcs->HCS_Base = iobase;

#else /* !__NetBSD__ */


void
a100_attach (pcici_t config_id, int unit)
{
	struct scsibus_data *scbus;
	u_int16_t io_port;
	u_int32_t id;
	UCHAR   irq;
	ULONG   wlval;
	A100_HCS    *pHcs;
	A100_SCB    *pScb;
	A100_SG 	*pSG;
	int     flags;

	if( unit >= MAX_SUPPORTED_ADAPTERS)
	return;


	io_port = 0;
	if (pci_map_port(config_id, PCI_BASEADDR0, &io_port) == 0)
	return;


	pHcs = (A100_HCS *) malloc (sizeof (A100_HCS), M_DEVBUF, M_NOWAIT);
	if( !pHcs ) {
		printf("ihb%d: cannot allocate HCS !\n", index);
		return;
	}

	bzero (pHcs, sizeof (A100_HCS));
	pHcs->HCS_Base = io_port;
	pHcs->HCS_Unit = unit;

	if (!(pci_map_int(config_id, (void *)orc_isr, (void *)pHcs, &bio_imask))) {
		printf("Can't pci_map_int \n");
		free(pHcs, M_DEVBUF);
		return;
	}


#endif /* !__NetBSD__ */

#if defined(__FreeBSD__)
	switch ((id = pci_conf_read(config_id, PCI_ID_REG))) {
#elif defined(__NetBSD__)
	switch (id = pa->pa_id) {
#endif

	    case PCI_DEVICE_ID_INICA10:
		break;

	    default:
		free(pHcs, M_DEVBUF);
		return;
	}


	pScb = (A100_SCB *) malloc (sizeof (A100_SCB)*MAX_SCB, M_DEVBUF, M_NOWAIT);
	if( !pScb ) {
		printf("ihb%d: cannot allocate SCB !\n", index);
		free(pHcs, M_DEVBUF);
		return;
	}
	bzero (pScb, sizeof (A100_SCB)*MAX_SCB);

	pSG = (A100_SG *) malloc (sizeof (A100_SG)*MAX_SCB*MAX_SG, M_DEVBUF, M_NOWAIT);
	if( !pSG ) {
		printf("ihb%d: cannot allocate SG !\n", index);
		free(pScb, M_DEVBUF);
		free(pHcs, M_DEVBUF);
		return;
	}
	orc_init(pHcs, pScb,pSG);
	/*
	 * fill in the prototype scsi_links.
	 */
#if defined(__FreeBSD__)
	pHcs->HCS_ScLink.adapter_unit = unit;
	pHcs->HCS_ScLink.adapter_targ = pHcs->HCS_SCSI_ID;
	pHcs->HCS_ScLink.fordriver = 0;
#elif defined(__NetBSD__)
	pHcs->HCS_ScLink.adapter_target = pHcs->HCS_SCSI_ID;
#endif
	pHcs->HCS_ScLink.adapter_softc = (void *)pHcs;
	pHcs->HCS_ScLink.adapter = &a100_switch;
	pHcs->HCS_ScLink.opennings = 1;
	pHcs->HCS_ScLink.device = &a100_dev;
	pHcs->HCS_ScLink.flags = DEBUGLEVEL;

#if defined(__FreeBSD__)
	/*
	 * Prepare the scsibus_data area for the upperlevel
	 * scsi code.
	 */
	scbus = scsi_alloc_bus();
	if(!scbus) 
	{
		free(pSG, M_DEVBUF);
		free(pScb, M_DEVBUF);
		free(pHcs, M_DEVBUF);
		return;
	}

	scbus->adapter_link = &pHcs->HCS_ScLink;
	scbus->maxtarg = pHcs->HCS_MaxTar - 1;

	/*
	 * ask the adapter what subunits are present
	 */
	scsi_attachdevs(scbus);

//	if (bootverbose)
		printf("ihb%d: Attach devices return\n", pHcs->HCS_Unit);

#elif defined(__NetBSD__)

	config_found((void *)pHcs, &pHcs->HCS_ScLink, a100print);

#endif
	return;
/*exit:			*/
/*	splx(flags);	*/

}



#ifdef __NetBSD__
int
a100_probe (parent, match, aux)
	struct device *parent;
	void *match, *aux;
{
	struct cfdata *cf = match;
	struct pci_attach_args *pa = aux;

	switch (pa->pa_id) {
	case PCI_DEVICE_ID_INICA10:
	    return 1;
	else
	    return (0);
	}
}

#else /* !__NetBSD__ */


static	char*
a100_probe (pcici_t tag, pcidi_t type)
{
	switch (type) {
	case PCI_DEVICE_ID_INICA10:
		return ("Initio INI-1060 PCI SCSI adapter");
	}
	return (NULL);
}

#endif /* !__NetBSD__ */



