/**************************************************************************
 * Copyright (c) 1997-99 Initio Corp.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *   Module: i91uscsi.c
 *   Description: PCI I/F for INIC-950, INIC-940i, & INIC-941 SCSI Bus Master Controller
 *   Revision History:
 *	11/18/96 HC, Port for tulip
 *	01/04/98 WH, Change to fit FreeBSD
 **********************************************************************/

#define DEBUG_INTERRUPT 0
#define DEBUG_QUEUE     0
#define DEBUG_STATE     0





#include <pci/i91uscsi.h>

/*--- external functions --*/
static void     tul_se2_wait(void);

#if 1
static void   tul_do_pause(ULONG amount );
#endif


/*--- forward refrence ---*/

static long      tulip_main(I91U_HCS *pHcs);

static I91U_SCB *tul_alloc_scb(I91U_HCS *pHcs);
static void tul_release_scb(I91U_HCS *pHcs, I91U_SCB *pScb);
static void tul_append_pend_scb(I91U_HCS *pHcs, I91U_SCB *scbp);
static void tul_push_pend_scb(I91U_HCS *pHcs, I91U_SCB *scbp);
static I91U_SCB *tul_find_pend_scb(I91U_HCS *pHcs);
static I91U_SCB *tul_pop_pend_scb(I91U_HCS *pHcs, I91U_SCB *pActScb);
static void tul_append_busy_scb(I91U_HCS *pHcs, I91U_SCB *scbp);
static void tul_unlink_busy_scb(I91U_HCS *pHcs, I91U_SCB *pScb);
static void tul_append_done_scb(I91U_HCS *pHcs, I91U_SCB *scbp);
static I91U_SCB *tul_find_done_scb(I91U_HCS *pHcs);
static I91U_SCB *tul_abort_ccb(I91U_HCS *pHcs, ULONG ccb);

static long      tul_next_state(I91U_HCS *pHcs);
static long      tul_state_5(I91U_HCS *pHcs);
static long      tul_xfer_data_in(I91U_HCS *pHcs);
static long      tul_xfer_data_out(I91U_HCS *pHcs);
static long      tul_xpad_in(I91U_HCS *pHcs);
static long      tul_xpad_out(I91U_HCS *pHcs);
static long      tul_status_msg(I91U_HCS *pHcs);
/*static long      tul_bad_seq(I91U_HCS *pHcs);	10/13/97	*/

static long      tul_msgin(I91U_HCS *pHcs);
static long      tul_msgin_sync(I91U_HCS *pHcs);
static long      tul_msgin_discon(I91U_HCS *pHcs);
static long      tul_msgin_accept(I91U_HCS *pHcs);
static long      tul_msgout_reject(I91U_HCS *pHcs);
static long      tul_msgin_extend(I91U_HCS *pHcs);
static long      tul_msgin_ignore_wid_resid(I91U_HCS *pHcs);

static	long      tul_msgout_sync(I91U_HCS *pHcs);
static	long      tul_msgout_nop(I91U_HCS *pHcs);
static	long      tul_msgout_ide(I91U_HCS *pHcs);
static	long      tul_msgout_abort_targ(I91U_HCS *pHcs);
static	long      tul_msgout_abort_tag(I91U_HCS *pHcs);

static	void     tul_select_atn(I91U_HCS *pHcs, I91U_SCB *pScb);
static	void     tul_select_atn3(I91U_HCS *pHcs, I91U_SCB *pScb);
static	void     tul_select_atn_stop(I91U_HCS *pHcs, I91U_SCB *pScb);
static	long      tul_reset_scsi(I91U_HCS *pHcs);
static	long      int_tul_busfree(I91U_HCS *pHcs);
static	long      int_tul_scsi_rst(I91U_HCS *pHcs);
static	long      int_tul_resel(I91U_HCS *pHcs);
static	void     tul_sync_done(I91U_HCS *pHcs);
static	void     tul_wdtr_done(I91U_HCS *pHcs);
static	long      wait_tulip(I91U_HCS *pHcs);
/*static long    tul_msgin_par_err(I91U_HCS *pHcs);*/
static	void     tulip_scsi(I91U_HCS *pHcs);
static	long      tul_post_scsi_rst(I91U_HCS *pHcs);

static	void     tul_se2_ew_en(DWORD port);
static	void     tul_se2_ew_ds(DWORD port);
static	long      tul_se2_rd_all(DWORD port);
static	void     tul_se2_update_all(DWORD port);       /* setup default pattern */
static	void     tul_read_eeprom(DWORD CurBase);


	/* ---- INTERNAL VARIABLES ---- */

/*NVRAM nvram, *nvramp = &nvram;*/
static NVRAM i91unvram;
static NVRAM *i91unvramp;



static UCHAR i91udftNvRam[64] = {               /*----------- header -----------*/
	0x25, 0xc9,             /* Signature    */
	0x40,                   /* Size         */
	0x01,                   /* Revision     */
	/* -- Host Adapter Structure -- */
	0x95,                   /* ModelByte0   */
	0x00,                   /* ModelByte1   */
	0x00,                   /* ModelInfo    */
	0x01,                   /* NumOfCh      */
	NBC1_DEFAULT,           /* BIOSConfig1  */
	0,                      /* BIOSConfig2  */
	0,                      /* HAConfig1    */
	0,                      /* HAConfig2    */
	/* SCSI channel 0 and target Structure  */
	7,                      /* SCSIid       */
	NCC1_DEFAULT,           /* SCSIconfig1  */
	0,                      /* SCSIconfig2  */
	8,                      /* NumSCSItarget*/

	NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT,
	NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT,
	NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT,
	NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT,

	/* SCSI channel 1 and target Structure  */
	7,                      /* SCSIid       */
	NCC1_DEFAULT,           /* SCSIconfig1  */
	0,                      /* SCSIconfig2  */
	8,                      /* NumSCSItarget*/

	NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT,
	NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT,
	NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT,
	NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT, NTC_DEFAULT,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0 };                         /*      - CheckSum -            */


static UCHAR    tul_rate_tbl[8] =       /* fast 20      */
		{
				/* nanosecond devide by 4 */
			12,     /* 50ns,  20M   */
			18,     /* 75ns,  13.3M */
			25,     /* 100ns, 10M   */
			31,     /* 125ns, 8M    */
			37,     /* 150ns, 6.6M  */
			43,     /* 175ns, 5.7M  */
			50,     /* 200ns, 5M    */
			62      /* 250ns, 4M    */
		};



#if 1
static void tul_do_pause( ULONG amount ) /* Pause for amount*1 milliseconds */
{
    ULONG i;

    for (i = 0; i < amount; i++)         /* wait      */
	DELAY (1000);  /* delay 1 msec */
}
#endif

/*-- forward reference --*/
#if 0
static void tul_snapdump(ULONG *ptr, int length)
{
    int col, row;
    int i, idx;

    row = length / 16 + ((length % 16) ? 1 : 0);
    idx = 0;
    printk("\n\r");
    for (i = 0; i < row; i++)
	{
	printf("Offset %04x at %08lx :  ", idx, (ULONG) ptr);
	for (col = 0; (col < 2) & (idx < length); col++, idx+=4)
	    {
	    printf(" %08lx", *ptr++);
	    }
	printk(" -");
	for (col = 0; (col < 2) & (idx < length); col++, idx+=4)
	    {
	    printk(" %08lx", *ptr++);
	    }
	printk("\n\r");
	}
}
#endif
/*--- forward reference ---*/

/*******************************************************************
	Use memeory refresh time        ~ 15us * 2
********************************************************************/
static void tul_se2_wait()
{
#if 1
	DELAY (5);  /* delay 5 usec */
#else
    UCHAR readByte;

    readByte = TUL_RD( 0x61 );
    if ( (readByte & 0x10) == 0x10) {
	for (; ;) {
	    readByte = TUL_RD( 0x61 );
	    if ( (readByte & 0x10) == 0x10)
		break;
	}
	for (;;) {
	    readByte = TUL_RD( 0x61 );
	    if ( (readByte & 0x10) != 0x10)
		break;
	}
    } else {
	for (; ; ) {
	    readByte = TUL_RD( 0x61 );
	    if ( (readByte & 0x10) == 0x10)
		break;
	}
	for (; ; ) {
	    readByte = TUL_RD( 0x61 );
	    if ( (readByte & 0x10) != 0x10)
		break;
	}
    }
#endif
}


/******************************************************************
 Input: instruction for  Serial E2PROM

******************************************************************/
static void tul_se2_instr(DWORD port, UCHAR instr)
{
    int i;
    UCHAR b;

    TUL_WR(port, SE2CS | SE2DO);     /* cs+start bit */
    tul_se2_wait();
    TUL_WR(port, SE2CS | SE2CLK | SE2DO); /* +CLK */
    tul_se2_wait();

    for ( i = 0; i < 8; i++) {
	if (instr & 0x80)
	    b =  SE2CS | SE2DO;         /* -CLK+dataBit */
	else
	    b =  SE2CS ;                /* -CLK */
	TUL_WR(port, b);
	tul_se2_wait();
	TUL_WR(port, b | SE2CLK);        /* +CLK */
	tul_se2_wait();
	instr <<= 1;
    }
    TUL_WR(port, SE2CS );                /* -CLK */
    tul_se2_wait();
    return;
}


/******************************************************************
 Function name  : tul_se2_ew_en
 Description    : Enable erase/write state of serial EEPROM
******************************************************************/
static void tul_se2_ew_en(DWORD port)
{
    tul_se2_instr(port, 0x30);                       /* EWEN */
    TUL_WR(port, 0);     /* -CS  */
    tul_se2_wait();
}


/************************************************************************
 Disable erase/write state of serial EEPROM
*************************************************************************/
static void tul_se2_ew_ds(DWORD port)
{
    tul_se2_instr(port,  0);         /* EWDS */
    TUL_WR(port, 0);     /* -CS  */
    tul_se2_wait();
}


/******************************************************************
	Input  :address of Serial E2PROM
	Output :value stored in  Serial E2PROM
*******************************************************************/
static USHORT tul_se2_rd(DWORD port, ULONG adr)
{
    UCHAR instr, readByte;
    USHORT readWord;
    int i;

    instr = (UCHAR)(  adr  | 0x80 );
    tul_se2_instr(port, instr);                       /* READ INSTR */
    readWord = 0;

    for ( i = 15; i >= 0; i--) {
	TUL_WR(port, SE2CS | SE2CLK );    /* +CLK */
	tul_se2_wait();
	TUL_WR(port, SE2CS );             /* -CLK */

		/* sample data after the following edge of clock  */
	readByte = TUL_RD(port);
	readByte &= SE2DI;
	readWord += (readByte << i);
	tul_se2_wait();                                        /* 6/20/95 */
    }

    TUL_WR(port, 0 );    /* no chip select */
    tul_se2_wait();
    return readWord;
}


/******************************************************************
 Input: new value in  Serial E2PROM, address of Serial E2PROM
*******************************************************************/
static void tul_se2_wr(DWORD port, UCHAR adr, USHORT writeWord)
{
    UCHAR readByte;
    UCHAR instr;
    int i;

    instr = (UCHAR)(  adr  | 0x40 );
    tul_se2_instr(port, instr);                      /* WRITE INSTR */
    for ( i = 15; i >= 0; i--) {
	if (writeWord & 0x8000)
	    TUL_WR(port , SE2CS | SE2DO); /* -CLK+dataBit 1 */
	else
	    TUL_WR(port , SE2CS );                  /* -CLK+dataBit 0 */
	tul_se2_wait();
	TUL_WR(port , SE2CS | SE2CLK );   /* +CLK */
	tul_se2_wait();
	writeWord <<= 1;
    }
    TUL_WR(port , SE2CS); /* -CLK */
    tul_se2_wait();
    TUL_WR(port , 0);     /* -CS  */
    tul_se2_wait();

    TUL_WR(port , SE2CS); /* +CS  */
    tul_se2_wait();

    for (; ; ) {
	TUL_WR(port, SE2CS | SE2CLK);       /* +CLK */
	tul_se2_wait();
	TUL_WR(port, SE2CS );               /* -CLK */
	tul_se2_wait();
	if ((readByte = TUL_RD(port)) & SE2DI )
	    break;                      /* write complete */
    }
    TUL_WR(port, 0);     /* -CS */
    return;
}


/***********************************************************************
 Read SCSI H/A configuration parameters from serial EEPROM
************************************************************************/
static long tul_se2_rd_all(DWORD port)
{
    int i;
    ULONG chksum = 0;
    USHORT *np;

    i91unvramp = &i91unvram;
    np = (USHORT * )i91unvramp;
    for ( i = 0; i < 32; i++) {
	*np++ = tul_se2_rd(port, i);
    }

	/*--------------------Is signature "ini" ok ? ----------------*/
    if ( i91unvramp->NVM_Signature != INI_SIGNATURE )
	return - 1;
#if 0
	/*--------------------Is Model# byte 1 "91" ok ? ----------------*/
//    if ( i91unvramp->ModelByte0 != 0x91)      /* 4:   Model number (byte 0)  */
//      return - 1;
#endif
	/*---------------------- Is ckecksum ok ? ----------------------*/
    np = (USHORT * )i91unvramp;
    for ( i = 0; i < 31; i++)
	chksum += *np++;
    if ( i91unvramp->NVM_CheckSum != (USHORT)chksum )
	return - 1;
    return 1;
}


/***********************************************************************
 Update SCSI H/A configuration parameters from serial EEPROM
************************************************************************/
static void tul_se2_update_all(DWORD port)   /* setup default pattern */
{
    int i;
    ULONG chksum = 0;
    USHORT * np, *np1;

    i91unvramp = &i91unvram;
	/* Calculate checksum first */
    np = (USHORT * )i91udftNvRam;
    for ( i = 0; i < 31; i++)
	chksum += *np++;
    *np = (USHORT)chksum;
    tul_se2_ew_en(port);                     /* Enable write  */

    np = (USHORT * )i91udftNvRam;
    np1 = (USHORT * )i91unvramp;
    for ( i = 0; i < 32; i++, np++, np1++) {
	if (*np != *np1) {
	    tul_se2_wr(port, i, *np);
	}
    }

    tul_se2_ew_ds(port);                     /* Disable write   */
    return;
}

/*************************************************************************
 Function name  : read_eeprom
**************************************************************************/
static void tul_read_eeprom(DWORD CurBase)
{
    UCHAR gctrl;

    i91unvramp = &i91unvram;
	/*------Enable EEProm programming ---*/
    gctrl = TUL_RD(CurBase + TUL_GCTRL);
    TUL_WR(CurBase + TUL_GCTRL, gctrl | TUL_GCTRL_EEPROM_BIT );

    if ( tul_se2_rd_all(CurBase+ TUL_NVRAM) != 1 )  {
	tul_se2_update_all(CurBase+ TUL_NVRAM);    /* setup default pattern */
	tul_se2_rd_all(CurBase+ TUL_NVRAM);                 /* load again  */
    }

	/*------ Disable EEProm programming ---*/
    gctrl = TUL_RD(CurBase + TUL_GCTRL);
    TUL_WR(CurBase + TUL_GCTRL, gctrl & ~TUL_GCTRL_EEPROM_BIT );
} /* read_eeprom */

static void tul_stop_bm(I91U_HCS *pHcs)
{

    if (TUL_RD(pHcs->HCS_Base + TUL_XStatus) & XPEND)
	{                       /* if DMA xfer is pending, abort DMA xfer */
	TUL_WR(pHcs->HCS_Base+ TUL_XCmd, TAX_X_ABT );
				/* wait Abort DMA xfer done */
	while ((TUL_RD(pHcs->HCS_Base + TUL_Int) & XABT) == 0);
	}
    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
}


/***************************************************************************/
static long init_tulip(I91U_HCS *pHcs, I91U_SCB *scbp, int num_scb)
{
    int         i;
    WORD        *pwFlags;
    I91U_SCB    *pTmpScb;
    I91U_SCB    *pPrevScb = NULL;
    TCS *pTcb;

    pHcs->HCS_SFifo = pHcs->HCS_Base + TUL_SFifo;
    pHcs->HCS_SCmd  = pHcs->HCS_Base + TUL_SCmd;

    pHcs->HCS_NumScbs = num_scb;
    pHcs->HCS_Semaph = 1;
    pHcs->HCS_JSStatus0 = 0;
    pHcs->HCS_Scb = scbp;
    for (i = 0, pTmpScb = scbp; i < num_scb; i++, pTmpScb++) {
	pTmpScb->SCB_TagId = i;
	pTmpScb->SCB_SGPAddr = (ULONG) vtophys(pTmpScb->SCB_SGList);
	if (i != 0)
	    pPrevScb->SCB_NxtScb = pTmpScb;
	pPrevScb = pTmpScb;
    }
    pPrevScb->SCB_NxtScb = NULL;


    pHcs->HCS_FirstAvail = scbp;
    pHcs->HCS_LastAvail = pPrevScb;

    pHcs->HCS_FirstPend = NULL;
    pHcs->HCS_LastPend = NULL;
    pHcs->HCS_FirstDone = NULL;
    pHcs->HCS_LastDone = NULL;
    pHcs->HCS_ActScb = NULL;
    pHcs->HCS_ActTcs = NULL;

    pHcs->HCS_Intr = TUL_RD(pHcs->HCS_Base + TUL_PINTL);

    tul_read_eeprom( pHcs->HCS_Base );
			/*---------- get H/A configuration -------------*/
    if (i91unvramp->NVM_SCSIInfo[0].NVM_NumOfTarg == 16)
	pHcs->HCS_MaxTar = 16;
    else
	pHcs->HCS_MaxTar = 8;

    pHcs->HCS_Config = i91unvramp->NVM_SCSIInfo[0].NVM_ChConfig1;

    pHcs->HCS_SCSI_ID = i91unvramp->NVM_SCSIInfo[0].NVM_ChSCSIID;

#if CHK_PARITY
					/* Enable parity error response */
    TUL_WR(pHcs->HCS_Base + TUL_PCMD, TUL_RD(pHcs->HCS_Base + TUL_PCMD) | 0x40);
#endif

					/* Mask all the interrupt       */
    TUL_WR(pHcs->HCS_Base + TUL_Mask, 0x1F);

    tul_stop_bm(pHcs);
					/* --- Initialize the tulip --- */
    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_RST_CHIP);

					/* program HBA's SCSI ID        */
    TUL_WR(pHcs->HCS_Base + TUL_SScsiId, pHcs->HCS_SCSI_ID << 4);

	/* Enable Initiator Mode ,phase latch,alternate sync period mode,
	disable SCSI reset */
    if (pHcs->HCS_Config & HCC_EN_PAR)
	pHcs->HCS_SConf1 = (TSC_INITDEFAULT | TSC_EN_SCSI_PAR);
    else
	pHcs->HCS_SConf1 = (TSC_INITDEFAULT);
    TUL_WR(pHcs->HCS_Base + TUL_SConfig, pHcs->HCS_SConf1);


	/* selection time out = 250 ms */
    TUL_WR(pHcs->HCS_Base + TUL_STimeOut, 153);

	/*--------- Enable SCSI terminator -----*/
    TUL_WR(pHcs->HCS_Base + TUL_XCtrl, (pHcs->HCS_Config & (HCC_ACT_TERM1 | HCC_ACT_TERM2)));
    TUL_WR(pHcs->HCS_Base + TUL_GCTRL1,
	  ((pHcs->HCS_Config & HCC_AUTO_TERM) >> 4) | (TUL_RD(pHcs->HCS_Base + TUL_GCTRL1) & 0xFE));

    pTcb = &pHcs->HCS_Tcs[0];
    for (i = 0,
	 pwFlags = (WORD *) &(i91unvramp->NVM_SCSIInfo[0].NVM_Targ0Config);
	 i < pHcs->HCS_MaxTar;
	 i++, pwFlags++, pTcb++ )
	{
	pTcb->TCS_Flags = *pwFlags & ~(TCF_SYNC_DONE | TCF_WDTR_DONE);
	pTcb->TCS_JS_Period = 0;
	pTcb->TCS_SConfig0 = pHcs->HCS_SConf1;
	pTcb->TCS_DrvFlags = 0;
	pTcb->TCS_TagCnt = 0;
	pTcb->TCS_NonTagScb = NULL;

	}                               /* for                          */
#if 0
    printf("iha%d: PCI Base=0x%x, IRQ=%d, SCSI ID=%d\n",
		pHcs->HCS_Unit, pHcs->HCS_Base + pHcs->HCS_Intr,
		pHcs->HCS_SCSI_ID);
#endif
	/*------------------- reset SCSI Bus ---------------------------*/
#if 1 
    printf("iha%d: Reset SCSI Bus ... \n", pHcs->HCS_Unit);
    tul_reset_scsi(pHcs);
#endif

    TUL_WR(pHcs->HCS_Base + TUL_SIntEnable, 0xFF);
    return(0);
}

/***************************************************************************/
static I91U_SCB *tul_alloc_scb(I91U_HCS *pHcs)
{
    I91U_SCB         *pTmpScb;
    ULONG       flags;

    flags = splcam();
    if ((pTmpScb = pHcs->HCS_FirstAvail) != NULL)
	{
#if DEBUG_QUEUE
	printf("find scb at %08lx\n", (ULONG) pTmpScb);
#endif
	if ((pHcs->HCS_FirstAvail = pTmpScb->SCB_NxtScb) == NULL)
	    pHcs->HCS_LastAvail = NULL;
	pTmpScb->SCB_NxtScb = NULL;
	pTmpScb->SCB_Status = SCB_RENT;
	}
    splx(flags);
    return (pTmpScb);
}

/***************************************************************************/
static void tul_release_scb(I91U_HCS *pHcs, I91U_SCB *pScb)
{
    ULONG       flags;

#if DEBUG_QUEUE
printf("Release SCB %lx; ", (ULONG) pScb);
#endif
    pScb->SCB_Status = 0;
    pScb->SCB_NxtScb = NULL;

    flags = splcam();
    if (pHcs->HCS_LastAvail != NULL) {
	pHcs->HCS_LastAvail->SCB_NxtScb = pScb;
	pHcs->HCS_LastAvail =pScb;
    }
    else {
	pHcs->HCS_FirstAvail = pScb;
	pHcs->HCS_LastAvail = pScb;
    }
    splx(flags);
}

/***************************************************************************/
static void tul_append_pend_scb(I91U_HCS *pHcs, I91U_SCB *scbp)
{
    ULONG       flags;

#if DEBUG_QUEUE
printf("Append pend SCB %lx; ", (ULONG) scbp);
#endif

    scbp->SCB_Status = SCB_PEND;
    scbp->SCB_NxtScb = NULL;

    flags = splcam();
    if (pHcs->HCS_LastPend != NULL) {
	pHcs->HCS_LastPend->SCB_NxtScb = scbp;
	pHcs->HCS_LastPend = scbp;
    }
    else {
	pHcs->HCS_FirstPend = scbp;
	pHcs->HCS_LastPend = scbp;
    }
    splx(flags);
}

/***************************************************************************/
static void tul_push_pend_scb(I91U_HCS *pHcs, I91U_SCB *scbp)
{
    ULONG       flags;

#if DEBUG_QUEUE
printf("Push pend SCB %lx; ", (ULONG) scbp);
#endif

    scbp->SCB_Status = SCB_PEND;

    flags = splcam();

    if ((scbp->SCB_NxtScb = pHcs->HCS_FirstPend) != NULL) {
	pHcs->HCS_FirstPend = scbp;
    }
    else {
	pHcs->HCS_FirstPend = scbp;
	pHcs->HCS_LastPend = scbp;
    }
    splx(flags);
}

/***************************************************************************/
static I91U_SCB *tul_find_pend_scb(I91U_HCS *pHcs)
{
    I91U_SCB         *pScb;
    TCS         *pTcb;
    ULONG       flags;


    flags = splcam();

    pScb = pHcs->HCS_FirstPend;
    for ( ;pScb != NULL; pScb = pScb->SCB_NxtScb)
    {
       pTcb = &pHcs->HCS_Tcs[pScb->SCB_Target];
	
        if (pScb->SCB_Opcode == BusDevRst) {
                break;
	}
        if (pScb->SCB_TagMsg) {
	    if (pTcb->TCS_NonTagScb == NULL) {
		break;
	    }
        }
        else if (pScb->SCB_Mode == SCM_RSENS) {
                break;
        }
        else if (pScb->SCB_CDB[0] == 0x3) {
                break;
	}
	else  if (pTcb->TCS_TagCnt == 0) {
	    if (pTcb->TCS_NonTagScb == NULL) {
	        break;
	    }
	}
    } /* for */

    splx(flags);

    return (pScb);
}


/***************************************************************************/
static I91U_SCB *tul_pop_pend_scb(I91U_HCS *pHcs, I91U_SCB *pActScb)
{
    I91U_SCB         *pScb;
    I91U_SCB         *pNxtScb;
    ULONG       flags;


    flags = splcam();

    pNxtScb = pActScb->SCB_NxtScb;
    pActScb->SCB_NxtScb = NULL;

    pScb = pHcs->HCS_FirstPend; 
    if (pScb == pActScb) {
	if (pNxtScb == NULL)
	    pHcs->HCS_LastPend = NULL;
	pHcs->HCS_FirstPend = pNxtScb;
    }
    else {
        while (pScb) {
	    if (pScb->SCB_NxtScb == pActScb) {
		pScb->SCB_NxtScb = pNxtScb;
		if (pActScb == pHcs->HCS_LastPend)
		    pHcs->HCS_LastPend = pScb;
		pScb = pActScb;
		break;
                }
	    pScb = pScb->SCB_NxtScb;
        }
    }
    splx(flags);
#if DEBUG_QUEUE
printf("Pop pend SCB %lx; ", (ULONG) pScb);
#endif
    return (pScb);
}

/***************************************************************************/
static void tul_append_busy_scb(I91U_HCS *pHcs, I91U_SCB *scbp)
{
    TCS *pTcb = pHcs->HCS_ActTcs;

#if DEBUG_QUEUE
printf("append busy SCB %lx; ", (ULONG) scbp);
#endif
    scbp->SCB_Status = SCB_BUSY;
    scbp->SCB_NxtScb = NULL;

    if (scbp->SCB_TagMsg) {
	pTcb->TCS_TagCnt++;
    }
    else {
	pTcb->TCS_NonTagScb = scbp;
    }
}

/***************************************************************************/
static void tul_unlink_busy_scb(I91U_HCS *pHcs, I91U_SCB *pScb)
{
    TCS *pTcb = &pHcs->HCS_Tcs[pScb->SCB_Target];

#if DEBUG_QUEUE
printf("unlink busy SCBs=%x ", (ULONG) pScb);
#endif

    if (pScb->SCB_TagMsg) {
	if (pTcb->TCS_TagCnt)
		pTcb->TCS_TagCnt--;
    }
    else if (pTcb->TCS_NonTagScb == pScb) {
        pTcb->TCS_NonTagScb = 0;
    }

}

/***************************************************************************/
static void tul_append_done_scb(I91U_HCS *pHcs, I91U_SCB *scbp)
{
#if 0
    ULONG       flags;
#endif

#if DEBUG_QUEUE
printf("append done SCB %lx; ", (ULONG) scbp);
#endif

    scbp->SCB_Status = SCB_DONE;

#if 0
    flags = splcam();
#endif
    scbp->SCB_NxtScb = NULL;
    if (pHcs->HCS_LastDone != NULL)
	{
	pHcs->HCS_LastDone->SCB_NxtScb = scbp;
	pHcs->HCS_LastDone = scbp;
	}
    else
	{
	pHcs->HCS_FirstDone = scbp;
	pHcs->HCS_LastDone = scbp;
	}
/*    splx(flags); */
}

/***************************************************************************/
static I91U_SCB *tul_find_done_scb(I91U_HCS *pHcs)
{
    I91U_SCB         *pTmpScb;
#if 0

    ULONG       flags;

    flags = splcam();
#endif

    if ((pTmpScb = pHcs->HCS_FirstDone) != NULL)
	{
	if ((pHcs->HCS_FirstDone = pTmpScb->SCB_NxtScb) == NULL)
	    pHcs->HCS_LastDone = NULL;
	pTmpScb->SCB_NxtScb = NULL;
	}

#if 0
    splx(flags);
#endif

#if DEBUG_QUEUE
printf("find done SCB %lx; ", (ULONG) pTmpScb);
#endif
    return (pTmpScb);
}

/***************************************************************************/
static I91U_SCB *tul_abort_ccb(I91U_HCS *pHcs, ULONG ccb)
{
    I91U_SCB    *pScb, *pPrevScb;
    ULONG       i;
    int         flags;

    flags = splcam();

    pPrevScb = NULL;
    pScb = pHcs->HCS_FirstPend;   /* Check Pend queue */
    while (pScb != NULL) {
	if (pScb->SCB_ccb == ccb) {
	    if (pPrevScb == NULL) {
		if ((pHcs->HCS_FirstPend = pScb->SCB_NxtScb) == NULL)
		    pHcs->HCS_LastPend = NULL;
	    }
	    else {
		pPrevScb->SCB_NxtScb = pScb->SCB_NxtScb;
		if (pScb == pHcs->HCS_LastPend)
		    pHcs->HCS_LastPend = pPrevScb;
	    }

	    pScb->SCB_NxtScb = NULL;
	    tul_release_scb(pHcs, pScb);

            splx(flags);
	    return (pScb);
	}

	pPrevScb = pScb;
	pScb = pScb->SCB_NxtScb;
    }

    pScb = pHcs->HCS_Scb ;
    for (i = 0; i < pHcs->HCS_NumScbs; i++, pScb++) {
	if (pScb->SCB_Status & SCB_BUSY) 
	    if (pScb->SCB_ccb == ccb) {
	        tul_unlink_busy_scb(pHcs, pScb);

                tul_release_scb(pHcs, pScb);
                splx(flags);
                return (pScb);
	    } 
    }
    splx(flags);
    return (NULL);
}

/***************************************************************************/
static long     tul_bad_seq(I91U_HCS *pHcs)
{
    I91U_SCB *pScb;

    if ((pScb = pHcs->HCS_ActScb) != NULL)
	{

	tul_unlink_busy_scb(pHcs, pScb);

	pScb->SCB_HaStat = HOST_BAD_PHAS;
	pScb->SCB_TaStat = 0;

	tul_append_done_scb(pHcs, pScb);
	}

    tul_stop_bm(pHcs);

    tul_reset_scsi(pHcs);

    return (tul_post_scsi_rst(pHcs));
}

/************************************************************************/
static void tul_exec_scb(I91U_HCS *pHcs, I91U_SCB *pScb)
{
    ULONG       flags;

#if 0
printf("Enter tul_exec_scb %x %x ", pScb->SCB_Target, pScb->SCB_CDB[0]);
tul_do_pause(1000);
#endif

    pScb->SCB_Mode = 0;

    pScb->SCB_SGIdx = 0;
    pScb->SCB_SGMax = pScb->SCB_SGLen;

    if (pScb->SCB_Opcode  == BusDevRst) {	/* bus_dev_reset */

	tul_push_pend_scb(pHcs, pScb);    /* Insert this SCB to Pending queue */
    }
    else if (pScb->SCB_CDB[0] == 0x3) {	/* request sense command */

	pScb->SCB_Ident &= 0xBF;		/* no disconnect */

	tul_push_pend_scb(pHcs, pScb);    /* Insert this SCB to Pending queue */
    }
    else {

	tul_append_pend_scb(pHcs, pScb);      /* Append this SCB to Pending queue */
    }


    flags = splcam();
    if (pHcs->HCS_Semaph == 1) {
	    TUL_WR(pHcs->HCS_Base + TUL_Mask, 0x1F);
                 /* disable SCSI Int        */
	    pHcs->HCS_Semaph = 0;
            splx(flags);

	    tulip_main(pHcs);

	    flags = splcam();
	    pHcs->HCS_Semaph = 1;
            TUL_WR(pHcs->HCS_Base + TUL_Mask, 0x0F);
	           /* Enable SCSI Int       */
	}
    splx(flags);
}

/***************************************************************************/
static long tul_isr(I91U_HCS *pHcs)
{
    BYTE        stat;


    stat = TUL_RD(pHcs->HCS_Base + TUL_SStatus0);
#if DEBUG_INTERRUPT
    printf("SCSI status = %x", stat);
#endif
    if (stat & TSS_INT_PENDING) {
	if (pHcs->HCS_Semaph == 1) {
            TUL_WR(pHcs->HCS_Base + TUL_Mask, 0x1F); /* Disable SCSI Int */
	    pHcs->HCS_Semaph = 0;

	    tulip_main(pHcs);

	    pHcs->HCS_Semaph = 1;
            TUL_WR(pHcs->HCS_Base + TUL_Mask, 0x0F);
	           /* Enable SCSI Int       */

	    return 1;
	}
	return 0; /* 08/11/95 */
    }
    return 0; /* 08/11/95 */
}

/***************************************************************************/
static long tulip_main(I91U_HCS *pHcs)
{
    I91U_SCB *pScb;
    TCS *pTcb;

/*printf("Enter tulip_main\n");*/
    for (;;) {
	tulip_scsi(pHcs);            /* Call tulip_scsi              */
	while ((pScb = tul_find_done_scb(pHcs)) != NULL)
	    {/* find done entry */
	    if (!(pScb->SCB_Mode & SCM_RSENS))
		{                       /* not in auto req. sense mode*/
		if (pScb->SCB_TaStat == 2)
		    {
	            pTcb = &pHcs->HCS_Tcs[pScb->SCB_Target];
		    pTcb->TCS_Flags &= ~(TCF_SYNC_DONE | TCF_WDTR_DONE);  /* clr sync. nego flag */
		    if (pScb->SCB_Flags & SCF_SENSE)
			{
			pScb->SCB_BufLen = pScb->SCB_SenseLen;
			pScb->SCB_BufPtr = pScb->SCB_SensePtr;
			pScb->SCB_Flags &= ~(SCF_SG|SCF_DIR); /* for xfer_data_in */
		/* so, we won't report worng direction in xfer_data_in,
				and won't report HOST_DO_DU in state_6 */
	                pScb->SCB_Ident &= 0xBF;	/* disconnect disable */
                        pScb->SCB_TagMsg = 0;        /* non-tag */
			pScb->SCB_Mode = SCM_RSENS;
			pScb->SCB_TaStat = 0;
			pScb->SCB_CDBLen = 6;
			pScb->SCB_CDB[0] = SCSICMD_RequestSense;
			pScb->SCB_CDB[1] = 0;
			pScb->SCB_CDB[2] = 0;
			pScb->SCB_CDB[3] = 0;
			pScb->SCB_CDB[4] = pScb->SCB_SenseLen;
			pScb->SCB_CDB[5] = 0;
			tul_push_pend_scb(pHcs, pScb);
			break;
			}
		    }
		}
	    else
		{                       /* in request sense mode */
		if (pScb->SCB_TaStat == 2)
		    {              /* check contition status again after sending
					requset sense cmd 0x3*/
		    pScb->SCB_HaStat = HOST_BAD_PHAS;
		    }
		pScb->SCB_TaStat = 2;
		}
	    pScb->SCB_Flags |= SCF_DONE;
	    if (pScb->SCB_Flags & SCF_POST)
		(*pScb->SCB_Post)((BYTE *) pHcs, (BYTE *) pScb);
	    }                           /* while */
					/* find_active: */
	if (TUL_RD(pHcs->HCS_Base + TUL_SStatus0) & TSS_INT_PENDING)
	    continue;

	if (pHcs->HCS_ActScb)        /* return to OS and wait for xfer_done_ISR/Selected_ISR */
	    return 1; /* return to OS, enable interrupt */
					/* Check pending I91U_SCB            */
	if (tul_find_pend_scb(pHcs) == NULL) {
	    return 1;                   /* return to OS, enable interrupt */
	}
    }       /* End of for loop */
	/* statement won't reach here */
}





/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/***************************************************************************/
/***************************************************************************/
/***************************************************************************/
/***************************************************************************/

/***************************************************************************/
static void tulip_scsi(I91U_HCS *pHcs)
{
    I91U_SCB         *pScb;
    TCS         *pTcb;
    BYTE        stat;

/*    printf("Enter tulip_scsi()\n");*/
	/* make sure to service interrupt asap */
    if ((stat = TUL_RD(pHcs->HCS_Base + TUL_SStatus0)) & TSS_INT_PENDING) {
	pHcs->HCS_JSStatus0 = stat;
	pHcs->HCS_Phase = stat & TSS_PH_MASK;
	pHcs->HCS_JSStatus1 = TUL_RD(pHcs->HCS_Base + TUL_SStatus1);
	pHcs->HCS_JSInt = TUL_RD(pHcs->HCS_Base + TUL_SInt);
/*    printf("SCSI status = %x\n", stat);
    printf("SCSI status 1 = %x\n", pHcs->HCS_JSStatus1);
    printf("SCSI interrupt = %x\n", pHcs->HCS_JSInt);*/
	if (pHcs->HCS_JSInt & TSS_SCSIRST_INT)
	    {                           /* SCSI bus reset detected      */
	    int_tul_scsi_rst(pHcs);
	    return;
	    }
	if (pHcs->HCS_JSInt & TSS_RESEL_INT) {      /* reselected interrupt */
	    int_tul_resel(pHcs);
	    return;
	}

	if (pHcs->HCS_JSInt & TSS_SEL_TIMEOUT) {
	    int_tul_busfree(pHcs);
	    return;
	}

	if (pHcs->HCS_JSInt & TSS_DISC_INT) {        /* BUS disconnection   */
	    int_tul_busfree(pHcs); /* unexpected bus free or sel timeout */
	    return;
	}

	if (pHcs->HCS_JSInt & (TSS_FUNC_COMP | TSS_BUS_SERV)) {                           /* func complete or Bus service */
	    if ((pScb = pHcs->HCS_ActScb) != NULL)
		tul_next_state(pHcs);
	    return;
	}

	if (pHcs->HCS_JSInt & TSS_SEL_INT) {     /*  */
		TUL_WR(pHcs->HCS_Base + TUL_SSignal, 0);
	}
    }

    if (pHcs->HCS_ActScb != NULL)
	return;

    if ((pScb = tul_find_pend_scb(pHcs)) == NULL)
	return;

    pTcb = &pHcs->HCS_Tcs[pScb->SCB_Target];
			/* program HBA's SCSI ID & target SCSI ID */    
    pHcs->HCS_ActTcs = pTcb ;
    TUL_WR( pHcs->HCS_Base + TUL_SScsiId,
			(pHcs->HCS_SCSI_ID << 4) | (pScb->SCB_Target & 0x0F) );
    if (pScb->SCB_Opcode == ExecSCSI) {

	TUL_WR(pHcs->HCS_Base + TUL_SPeriod, pTcb->TCS_JS_Period);
/*printf("TCS_Flags = %x\n", pTcb->TCS_Flags);
do_pause(15000);*/
	if ((pTcb->TCS_Flags & (TCF_WDTR_DONE | TCF_NO_WDTR)) == 0) {                           /* do wdtr negotiation          */
	    tul_select_atn_stop(pHcs, pScb);
	}
	else if ((pTcb->TCS_Flags & (TCF_SYNC_DONE | TCF_NO_SYNC_NEGO)) == 0) {                       /* do sync negotiation          */
	    tul_select_atn_stop(pHcs, pScb);
	}
	else if (pScb->SCB_TagMsg)
	    tul_select_atn3(pHcs, pScb);
	else
	    tul_select_atn(pHcs, pScb);

	if (pScb->SCB_Flags & SCF_POLL) {
	    while (wait_tulip(pHcs) != -1) {
		if (tul_next_state(pHcs) == -1)
		    break;
	    }
	}
    }
    else if (pScb->SCB_Opcode == BusDevRst) {
	tul_select_atn_stop(pHcs, pScb);
	pScb->SCB_NxtStat = 8;
	if (pScb->SCB_Flags & SCF_POLL) {
		while (wait_tulip(pHcs) != -1) {
		    if (tul_next_state(pHcs) == -1)
			break;
		    }
		}
    }
    else {
        tul_pop_pend_scb(pHcs, pScb);

	pScb->SCB_HaStat = HOST_INV_CMD    ;    /* bad command */
	tul_append_done_scb(pHcs, pScb);
    }
}


/***************************************************************************/
static long     tul_next_state(I91U_HCS *pHcs)
{
    I91U_SCB    *pScb = pHcs->HCS_ActScb;
    TCS         *pTcb = pHcs->HCS_ActTcs;
    int         next = pScb->SCB_NxtStat;
#if 0
    int i;
#endif

    switch (next) {

	case 1: {
	    if (pScb != tul_pop_pend_scb(pHcs, pScb)) {
		printf("iha:Differnet SCB state 1\n");
		}
	    tul_append_busy_scb(pHcs, pScb);

	    TUL_WR(pHcs->HCS_Base + TUL_SConfig, pTcb->TCS_SConfig0);
					/* ATN on */
	    if (pHcs->HCS_Phase == MSG_OUT) {
		TUL_WR(pHcs->HCS_Base+TUL_SCtrl1, (TSC_EN_BUS_IN|TSC_HW_RESELECT));

		TUL_WR(pHcs->HCS_SFifo, pScb->SCB_Ident);

		if (pScb->SCB_TagMsg) {
		    TUL_WR(pHcs->HCS_SFifo, pScb->SCB_TagMsg);
		    TUL_WR(pHcs->HCS_SFifo, pScb->SCB_TagId);
		}

		if ((pTcb->TCS_Flags & (TCF_WDTR_DONE | TCF_NO_WDTR)) == 0) {
		    pTcb->TCS_Flags |= TCF_WDTR_DONE;
		    TUL_WR(pHcs->HCS_SFifo, MSG_EXTEND);
		    TUL_WR(pHcs->HCS_SFifo, 2);     /* Extended message length*/
		    TUL_WR(pHcs->HCS_SFifo, 3);     /* Sync request*/
		    TUL_WR(pHcs->HCS_SFifo, 1);     /* Start from 16 bits xfer*/
		    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
		    if (wait_tulip(pHcs) == -1)
			return(-1);
		    else {
			TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
			TUL_WR(pHcs->HCS_Base + TUL_SSignal, (TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7)));
		        goto state_3;
		    }
		}
		else {
		    if ((pTcb->TCS_Flags & (TCF_SYNC_DONE | TCF_NO_SYNC_NEGO)) == 0) {
			if (tul_msgout_sync(pHcs) == -1)
			    return(-1);
		    }
		    goto state_3;
		}
    	    }
	    else {
		TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
		TUL_WR(pHcs->HCS_Base + TUL_SSignal, (TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7)));
		goto state_3;
	    }
	}

	case 2: {
#if DEBUG_STATE
printf("-s2-");
#endif
	    if (pScb != tul_pop_pend_scb(pHcs, pScb)) {
		printf("iha: Differnet SCB state 2\n");
	    }
/*	    pScb->SCB_Status = SCB_BUSY;		*/
	    tul_append_busy_scb(pHcs, pScb);

	    TUL_WR(pHcs->HCS_Base + TUL_SConfig, pTcb->TCS_SConfig0);

	    if ( pHcs->HCS_JSStatus1 & TSS_CMD_PH_CMP) {
		goto state_4;
	    }
	    /* clear ATN */
	    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
	    TUL_WR(pHcs->HCS_Base + TUL_SSignal, (TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7)));
/*	    goto state_3;		*/
	}

	case 3: {
state_3:

#if DEBUG_STATE
print("-s3-");
#endif
	    for (;;) {
		switch (pHcs->HCS_Phase) {

		    case CMD_OUT:            /* Command out phase            */
#if 0
			for ( i = 0; i < (int) pScb->SCB_CDBLen; i++)
			    TUL_WR(pHcs->HCS_SFifo, pScb->SCB_CDB[i]);
#else
			outsb(pHcs->HCS_SFifo, pScb->SCB_CDB, pScb->SCB_CDBLen);
#endif
			TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
			if (wait_tulip(pHcs) == -1)
			    return (-1);
			if (pHcs->HCS_Phase == CMD_OUT) {
			    return (tul_bad_seq(pHcs));
			}
			goto state_4;

		    case MSG_IN:            /* Message in phase             */
			pScb->SCB_NxtStat = 3;
			if (tul_msgin(pHcs) == -1)
			    return (-1);
			break;

		    case STATUS_IN:         /* Status phase                 */
			if (tul_status_msg(pHcs) == -1)
			    return (-1);
			break;

		    case MSG_OUT:           /* Message out phase            */
			if (pTcb->TCS_Flags & (TCF_SYNC_DONE|TCF_NO_SYNC_NEGO)) {
			    if (tul_msgout_nop(pHcs) == -1)
				return (-1);
			} else {
			    if (tul_msgout_sync(pHcs) == -1)
				return(-1);
			}
			break;

		    default:
			return (tul_bad_seq(pHcs));
		}
	    }
	}

	case 4:
state_4:
#if DEBUG_STATE
print("-s4-");
#endif
	    if ((pScb->SCB_Flags & SCF_DIR) == SCF_NO_XF) {
		goto state_6;         /* Go to state 6                */
	    }

	    for (;;) {
		if (pScb->SCB_BufLen == 0)
			goto state_6;         /* Go to state 6                */
		switch (pHcs->HCS_Phase) {
		    case STATUS_IN:           /* Status phase                 */
			if ((pScb->SCB_Flags & SCF_DIR) != 0) {
			 /* if direction bit set then report data underrun */
			    pScb->SCB_HaStat = HOST_DO_DU;
			}
			if ((tul_status_msg(pHcs)) == -1)
			    return (-1);
			break;
		    case MSG_IN:             /* Message in phase             */
			pScb->SCB_NxtStat = 4;
			if (tul_msgin(pHcs) == -1)
			    return (-1);
			break;
		    case MSG_OUT:             /* Message out phase            */
			if (pHcs->HCS_JSStatus0 & TSS_PAR_ERROR) {
			    pScb->SCB_BufLen = 0;
			    pScb->SCB_HaStat = HOST_DO_DU;
			    if (tul_msgout_ide(pHcs) == -1)
				return (-1);
			    goto state_6;         /* Go to state 6           */
			} else {
			    if (tul_msgout_nop(pHcs) == -1)
				return (-1);
			}
			break;

		    case DATA_IN:               /* Data in phase             */
			return (tul_xfer_data_in(pHcs));

		    case DATA_OUT:              /* Data out phase            */
			return (tul_xfer_data_out(pHcs));
		    case CMD_OUT:               /* Command out phase         */
		    default:
			return (tul_bad_seq(pHcs));
		}
	    }

	case 5:
		if ((next = tul_state_5(pHcs)) == -1)
		    return -1;
		if (next == 4)
		    goto state_4;

	case 6:
state_6:
#if DEBUG_STATE
print("-s6-");
#endif
	    for (;;) {
		switch (pHcs->HCS_Phase) {
		    case STATUS_IN:           /* Status phase                 */
			if (tul_status_msg(pHcs) == -1)
			    return (-1);
			break;
		    case MSG_IN:              /* Message in phase             */
			pScb->SCB_NxtStat = 6;
			if ((tul_msgin(pHcs)) == -1)
			    return (-1);
			break;
		    case MSG_OUT:             /* Message out phase            */
			if ((tul_msgout_nop(pHcs)) == -1)
			    return (-1);
			break;
		    case DATA_IN:             /* Data in phase                */
			if (tul_xpad_in(pHcs) == -1)
			    return (-1);
			break;
		    case DATA_OUT:            /* Data out phase               */
			if (tul_xpad_out(pHcs) == -1)
			    return (-1);
			break;
		    default:
			return (tul_bad_seq(pHcs));
		}
	    }

	case 8: {
	    ULONG       i;
	    BYTE        tar;

	    if (pScb != tul_pop_pend_scb(pHcs, pScb)) {
		printf("Differnet SCB state 1\n");
	    }

	    if (pHcs->HCS_Phase != MSG_OUT) {
		return (tul_bad_seq(pHcs));      /* Unexpected phase         */
	    }

	    TUL_WR(pHcs->HCS_SFifo, MSG_DEVRST);
	    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);

	    pScb->SCB_HaStat = 0;
	    tul_append_done_scb(pHcs, pScb);

	    pTcb->TCS_Flags &= ~(TCF_SYNC_DONE | TCF_WDTR_DONE);
	    pTcb->TCS_JS_Period = 0;
	    pTcb->TCS_SConfig0 = pHcs->HCS_SConf1;
	    pTcb->TCS_TagCnt = 0;
	    pTcb->TCS_NonTagScb = NULL;

		/* abort all SCB with same target */
	    tar = pScb->SCB_Target;          /* target                       */

	    pScb = pHcs->HCS_Scb ;
	    for (i = 0; i < pHcs->HCS_NumScbs; i++, pScb++) {
		if ((pScb->SCB_Status & SCB_BUSY) &&
		    (pScb->SCB_Target == tar) ) {
		    tul_unlink_busy_scb(pHcs, pScb);

		    pScb->SCB_HaStat = HOST_DEV_RST;
		    tul_append_done_scb(pHcs, pScb);
	        } 
	    }

	    pHcs->HCS_Flags |=  HCF_EXPECT_DISC;
	    if (wait_tulip(pHcs) == -1) {
		    return (-1);
	    }
	    /* Unexpected phase         */
	}

	default:
		return (tul_bad_seq(pHcs));
    }
}



/***************************************************************************/
/* state after dma xfer done or phase change before xfer done */
static long     tul_state_5(I91U_HCS *pHcs)
{

    I91U_SCB    *pScb = pHcs->HCS_ActScb;
    TCS         *pTcb = pHcs->HCS_ActTcs;
    long        cnt, xcnt;  /* cannot use unsigned !! code: if (xcnt < 0) */

#if DEBUG_STATE
print("-s5-");
#endif

    /*------ get remaining count -------*/
    cnt = TUL_RDLONG(pHcs->HCS_Base + TUL_SCnt0) & 0x0FFFFFF;

/*print("Remaining SCSI Count = %lx\n", cnt);*/
    if (TUL_RD(pHcs->HCS_Base + TUL_XCmd) & 0x20)
	{
	/* ----------------------- DATA_IN -----------------------------*/
		/* check scsi parity error */
	if (pHcs->HCS_JSStatus0 & TSS_PAR_ERROR) {
	    pScb->SCB_HaStat = HOST_DO_DU;
	    }

	if (TUL_RD(pHcs->HCS_Base + TUL_XStatus) & XPEND)
	    {                           /* DMA xfer pending, Send STOP  */
			/* tell Hardware  scsi xfer has been terminated */
	    TUL_WR(pHcs->HCS_Base + TUL_XCtrl, TUL_RD(pHcs->HCS_Base + TUL_XCtrl) | 0x80);
			/* wait until DMA xfer not pending */
	    while (TUL_RD(pHcs->HCS_Base + TUL_XStatus) & XPEND );
	    }
	}
    else
	{                               /*-------- DATA OUT -----------*/
	if ((TUL_RD(pHcs->HCS_Base + TUL_SStatus1) & TSS_XFER_CMP) == 0)
	    {
	    if (pTcb->TCS_JS_Period & TSC_WIDE_SCSI)
		cnt += (TUL_RD(pHcs->HCS_Base + TUL_SFifoCnt) & 0x1F) << 1;
	    else
		cnt += (TUL_RD(pHcs->HCS_Base + TUL_SFifoCnt) & 0x1F);
	    }
	if (TUL_RD(pHcs->HCS_Base + TUL_XStatus) & XPEND)
	    {                   /* if DMA xfer is pending, abort DMA xfer */
	    TUL_WR(pHcs->HCS_Base + TUL_XCmd, TAX_X_ABT );
				/* wait Abort DMA xfer done */
	    while ((TUL_RD(pHcs->HCS_Base + TUL_Int) & XABT) == 0);
	    }
	if ((cnt == 1) && (pHcs->HCS_Phase == DATA_OUT)) {
	    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
	    if (wait_tulip(pHcs) == -1) {
		return (-1);
		}
	    cnt = 0;
	    }
	else
	    {
	    if ((TUL_RD(pHcs->HCS_Base + TUL_SStatus1) & TSS_XFER_CMP) == 0)
		TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
	    }
	}

    if (cnt == 0) {
	pScb->SCB_BufLen = 0;
	return (6);                     /* Go to state 6                */
	}
					/* Update active data pointer */
    xcnt = pScb->SCB_BufLen - cnt;   /* xcnt== bytes already xferred */
    pScb->SCB_BufLen = cnt;          /* cnt == bytes left to be xferred */
    if (pScb->SCB_Flags & SCF_SG)
	{
	register I91U_SG *sgp;
	ULONG       i;

	sgp = &pScb->SCB_SGList[pScb->SCB_SGIdx];
	for (i = pScb->SCB_SGIdx; i < pScb->SCB_SGMax; sgp++, i++)
	    {
	    xcnt -= sgp->SG_Len;
	    if (xcnt < 0)
		{               /* this sgp xfer half done */
		xcnt += sgp->SG_Len;   /* xcnt == bytes xferred in this sgp */
		sgp->SG_Ptr += (ULONG)xcnt;   /* new ptr to be xfer */
		sgp->SG_Len -= (ULONG)xcnt;   /* new len to be xfer */
		pScb->SCB_BufPtr += ((ULONG)(i - pScb->SCB_SGIdx) << 3);
					/* new SG table ptr */
		pScb->SCB_SGLen = (BYTE)(pScb->SCB_SGMax - i);
					/* new SG table len */
		pScb->SCB_SGIdx = (WORD)i;
					/* for next disc and come in this loop*/
		return (4);             /* Go to state 4                */
		}
			/* else (xcnt >= 0 , i.e. this sgp already xferred */
	    } /* for */
	return (6);                     /* Go to state 6                */
	}
     else {
	pScb->SCB_BufPtr += xcnt;
	}
    return (4);                         /* Go to state 4                */
}

/***************************************************************************/
static long     tul_xfer_data_in(I91U_HCS *pHcs)
{
    I91U_SCB *pScb = pHcs->HCS_ActScb;

/*print("\nEnter xfer data in; ptr = %lx; cnt = %lx\n", pScb->SCB_BufPtr,
							pScb->SCB_BufLen);*/
    if ((pScb->SCB_Flags & SCF_DIR) ==  SCF_DOUT)  {
	return (6);       /* wrong direction */
    }
    TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, pScb->SCB_BufLen);

    TUL_WR(pHcs->HCS_SCmd, TSC_XF_DMA_IN);          /* 7/25/95 */

    if (pScb->SCB_Flags & SCF_SG) {           /* S/G xfer */
	TUL_WRLONG(pHcs->HCS_Base + TUL_XCntH, ((ULONG)pScb->SCB_SGLen) << 3);
	TUL_WRLONG(pHcs->HCS_Base + TUL_XAddH, pScb->SCB_BufPtr);
	TUL_WR(pHcs->HCS_Base + TUL_XCmd, TAX_SG_IN);
    } else {
	TUL_WRLONG(pHcs->HCS_Base + TUL_XCntH, pScb->SCB_BufLen);
	TUL_WRLONG(pHcs->HCS_Base + TUL_XAddH, pScb->SCB_BufPtr);
	TUL_WR(pHcs->HCS_Base + TUL_XCmd, TAX_X_IN);
    }

    pScb->SCB_NxtStat = 5;
    return (0);  /* return to OS, wait xfer done , let jas_isr come in */
}


/***************************************************************************/
static long     tul_xfer_data_out(I91U_HCS *pHcs)
{
    I91U_SCB *pScb = pHcs->HCS_ActScb;

/*print("Enter xfer data out\n");*/
    if ((pScb->SCB_Flags & SCF_DIR) ==  SCF_DIN) {
	return (6);       /* wrong direction */
    }

    TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, pScb->SCB_BufLen);

    if (pScb->SCB_Flags & SCF_SG) {           /* S/G xfer */
	TUL_WRLONG(pHcs->HCS_Base + TUL_XCntH, ((ULONG)pScb->SCB_SGLen) << 3);
	TUL_WRLONG(pHcs->HCS_Base + TUL_XAddH, pScb->SCB_BufPtr);
	TUL_WR(pHcs->HCS_Base + TUL_XCmd, TAX_SG_OUT);
    } else {
	TUL_WRLONG(pHcs->HCS_Base + TUL_XCntH, pScb->SCB_BufLen);
	TUL_WRLONG(pHcs->HCS_Base + TUL_XAddH, pScb->SCB_BufPtr);
	TUL_WR(pHcs->HCS_Base + TUL_XCmd, TAX_X_OUT);
    }

    TUL_WR(pHcs->HCS_SCmd, TSC_XF_DMA_OUT);

    pScb->SCB_NxtStat = 5;
    return (0);  /* return to OS, wait xfer done , let jas_isr come in */
}


/***************************************************************************/
static long     tul_xpad_in(I91U_HCS *pHcs)
{
    I91U_SCB *pScb = pHcs->HCS_ActScb;
    TCS *pTcb = pHcs->HCS_ActTcs;
    int phase;

    if ((pScb->SCB_Flags & SCF_DIR) != SCF_NO_DCHK)  {
	pScb->SCB_HaStat = HOST_DO_DU;       /* over run             */
    }
    for (;;) {
	if (pTcb->TCS_JS_Period & TSC_WIDE_SCSI)
	    TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, 2);
	else
	    TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, 1);

	TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_IN);
	if ((phase = wait_tulip(pHcs)) == -1)
	    return (-1);

	if (phase != DATA_IN) {
	    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
	    return (6);
	    }
	TUL_RD(pHcs->HCS_SFifo);
    }
}

static long     tul_xpad_out(I91U_HCS *pHcs)
{
    I91U_SCB *pScb = pHcs->HCS_ActScb;
    TCS *pTcb = pHcs->HCS_ActTcs;
    int phase;

    if ((pScb->SCB_Flags & SCF_DIR) != SCF_NO_DCHK)  {
	pScb->SCB_HaStat = HOST_DO_DU;       /* over run             */
    }
    for (;;) {
	if (pTcb->TCS_JS_Period & TSC_WIDE_SCSI)
	    TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, 2);
	else
	    TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, 1);

	TUL_WR(pHcs->HCS_SFifo, 0);
	TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
	if ((phase = wait_tulip(pHcs)) == -1) 
	    return (-1);

	if (phase != DATA_OUT)
	    {                   /* Disable wide CPU to allow read 16 bits */
	    TUL_WR(pHcs->HCS_Base + TUL_SCtrl1, TSC_HW_RESELECT);
	    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
	    return (6);
	    }
    }
}


/***************************************************************************/
static long     tul_status_msg(I91U_HCS *pHcs)    /* status & MSG_IN */
{
    I91U_SCB    *pScb = pHcs->HCS_ActScb;
    char        phase;
    BYTE        msg;

    TUL_WR(pHcs->HCS_SCmd, TSC_CMD_COMP);
    if ((phase = wait_tulip(pHcs)) == -1) {
	return (-1);
    }
	/* get status */
    pScb->SCB_TaStat = TUL_RD(pHcs->HCS_SFifo);
#if 0
    phase = pHcs->HCS_Phase = TUL_RD(pHcs->HCS_Base + TUL_SSignal) & 0x07;
#endif

    if (phase == MSG_OUT) {
	if (pHcs->HCS_JSStatus0 & TSS_PAR_ERROR) {
	    TUL_WR(pHcs->HCS_SFifo, MSG_PARITY);
	}
	else {
	    TUL_WR(pHcs->HCS_SFifo, MSG_NOP);
	}
	TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
	return (wait_tulip(pHcs));
    }

    if (phase == MSG_IN) {
	msg = TUL_RD(pHcs->HCS_SFifo);
	if (pHcs->HCS_JSStatus0 & TSS_PAR_ERROR) {          /* Parity error   */
	    if ((phase = tul_msgin_accept(pHcs)) == -1)
		return (-1);
	    if (phase != MSG_OUT)
		return (tul_bad_seq(pHcs));
	    TUL_WR(pHcs->HCS_SFifo, MSG_PARITY);
	    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
	    return (wait_tulip(pHcs));
	}

	if (msg == 0) {                             /* Command complete      */
	    if ((pScb->SCB_TaStat & 0x18) == 0x10) {   /* No link support    */
		return (tul_bad_seq(pHcs));
	    }
	    pHcs->HCS_Flags |= HCF_EXPECT_DONE_DISC;
	    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
	    TUL_WR(pHcs->HCS_SCmd, TSC_MSG_ACCEPT);
	    return(wait_tulip(pHcs));
	}
	if ((msg ==  MSG_LINK_COMP) || (msg == MSG_LINK_FLAG))
	    {
	    if ((pScb->SCB_TaStat & 0x18) == 0x10)
		return (tul_msgin_accept(pHcs));
	    }
	}
    return (tul_bad_seq(pHcs));
}


/***************************************************************************/
/* scsi bus free */
static long     int_tul_busfree(I91U_HCS *pHcs)
{
    I91U_SCB *pScb = pHcs->HCS_ActScb;

    if (pScb != NULL)
	{
	if (pScb->SCB_Status & SCB_SELECT) { /* selection timeout */
	    if (pScb != tul_pop_pend_scb(pHcs, pScb)) {
		printf("iha: Differnet SCB\n");
		}
	    pScb->SCB_HaStat = HOST_SEL_TOUT;
	    tul_append_done_scb(pHcs, pScb);
	    }
	else {                           /* Unexpected bus free          */
	    tul_unlink_busy_scb(pHcs, pScb);

	    pScb->SCB_HaStat = HOST_BUS_FREE;
	    tul_append_done_scb(pHcs, pScb);
	    }
	pHcs->HCS_ActScb = NULL;
/*	pHcs->HCS_ActTcs = NULL;		*/
	}
    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO); /* Flush SCSI FIFO  */
    TUL_WR(pHcs->HCS_Base + TUL_SConfig, TSC_INITDEFAULT);
    TUL_WR(pHcs->HCS_Base + TUL_SCtrl1, TSC_HW_RESELECT);/* Enable HW reselect       */
    return (-1);
}


/***************************************************************************/
/* scsi bus reset */
static long     int_tul_scsi_rst(I91U_HCS *pHcs)
{
    I91U_SCB *pScb;
    int i;
    TCS *pTcb;

	/* if DMA xfer is pending, abort DMA xfer */
    if (TUL_RD(pHcs->HCS_Base + TUL_XStatus) & 0x01)
	{
	TUL_WR(pHcs->HCS_Base + TUL_XCmd, TAX_X_ABT );
		/* wait Abort DMA xfer done */
	while ((TUL_RD(pHcs->HCS_Base + TUL_Int) & 0x04) == 0)
			;
	TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
	}

	/* Abort all active & disconnected scb */
    pScb = pHcs->HCS_Scb ;
    for (i = 0; i < pHcs->HCS_NumScbs; i++, pScb++)
	{
	if (pScb->SCB_Status & SCB_BUSY )
	    {
	    tul_unlink_busy_scb(pHcs, pScb);

	    pScb->SCB_HaStat = HOST_SCSI_RST;
	    tul_append_done_scb(pHcs, pScb);
	    } 
	}

    pHcs->HCS_ActScb = NULL;

	/* clr sync nego. done flag */
    pTcb = &pHcs->HCS_Tcs[0];
    for (i = 0; i < pHcs->HCS_MaxTar; i++,pTcb++  )
	{
	pTcb->TCS_Flags &= ~(TCF_SYNC_DONE | TCF_WDTR_DONE);
	pTcb->TCS_JS_Period = 0;
	pTcb->TCS_SConfig0 = pHcs->HCS_SConf1;
	pTcb->TCS_TagCnt = 0;
	pTcb->TCS_NonTagScb = NULL;
	}
    return(-1);
}

#if 0
/***************************************************************************/
static long     tul_msgin_par_err(I91U_HCS *pHcs)
{
    int phase = 0;

    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
    if ((phase = tul_msgin_accept(pHcs)) == -1)
	return (-1);
    if (phase == MSG_OUT)
	{                               /* Assert attention             */
	TUL_WR(pHcs->HCS_Base + TUL_SSignal, ((TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7)) | TSC_SET_ATN));
	TUL_WR(pHcs->HCS_SFifo, MSG_PARITY);
	TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
	if ((phase = wait_tulip(pHcs)) == -1)
	    return (-1);
	if (phase == MSG_IN)
	    {
	    TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, 1);
	    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_IN);
	    if ((phase = wait_tulip(pHcs)) == -1)
		return (-1);
	    if ((pHcs->HCS_JSStatus0 & TSS_PAR_ERROR) == 0)
		return (phase);
	    }
	}
    return (tul_bad_seq(pHcs));
}
#endif

/***************************************************************************/
/* scsi reselection */
static long     int_tul_resel(I91U_HCS *pHcs)
{
    I91U_SCB    *pScb;
    TCS         *pTcb;
    int         phase;
    BYTE        tag;
    BYTE        msg = 0;
    WORD        tar, lun;

    if ((pScb = pHcs->HCS_ActScb) != NULL)
	{
	if (pScb->SCB_Status & SCB_SELECT)
	    {                           /* if waiting for selection complete*/
	    pScb->SCB_Status &= ~SCB_SELECT;
	    }
	pHcs->HCS_ActScb = NULL;
	}

    /* --------- get target id---------------------- */
    tar =  TUL_RD(pHcs->HCS_Base + TUL_SBusId);
    /* ------ get LUN from Identify message----------- */
    lun = TUL_RD(pHcs->HCS_Base + TUL_SIdent) & 0x1F;	/* 10/21/97 from 0x07 -> 0x1f */

    pTcb = &pHcs->HCS_Tcs[tar];
    pHcs->HCS_ActTcs = pTcb;
    TUL_WR(pHcs->HCS_Base + TUL_SConfig, pTcb->TCS_SConfig0);
    TUL_WR(pHcs->HCS_Base + TUL_SPeriod, pTcb->TCS_JS_Period);


    /* ------------- tag queueing ? ------------------- */
    if (pTcb->TCS_NonTagScb == NULL) {
/*printf("Tag enabled\n");*/
	if ((phase = tul_msgin_accept(pHcs)) == -1)
	    return (-1);
	if (phase != MSG_IN)
	    goto no_tag;
	TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, 1);
	TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_IN);
	if ((phase = wait_tulip(pHcs)) == -1)
	    return (-1);
	msg = TUL_RD(pHcs->HCS_SFifo);       /* Read Tag Message     */

	if ((msg < MSG_STAG) || (msg > MSG_OTAG)) /* Is simple Tag      */
	    goto no_tag;

	if ((phase = tul_msgin_accept(pHcs)) == -1)
	    return (-1);

	if (phase != MSG_IN)
	    goto no_tag;

	TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, 1);
	TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_IN);
	if ((phase = wait_tulip(pHcs)) == -1)
	    return (-1);
	tag = TUL_RD(pHcs->HCS_SFifo);       /* Read Tag ID          */
	pScb = pHcs->HCS_Scb + tag;
	if ((pScb->SCB_Target != tar)  || (pScb->SCB_Lun != lun))
	    {
	    return tul_msgout_abort_tag(pHcs);
	    }
	if (pScb->SCB_Status != SCB_BUSY)
	    {                                   /* 03/24/95             */
	    return tul_msgout_abort_tag(pHcs);
	    }
	pHcs->HCS_ActScb = pScb;
	if ((phase = tul_msgin_accept(pHcs)) == -1)
	    return (-1);
	}
    else                                        /* No tag               */
	{
/*printf("tarlun= %x\n", tarlun);*/
        pScb = pTcb->TCS_NonTagScb; 
	if ((pScb->SCB_Status != SCB_BUSY)|| (pScb->SCB_Lun != lun))
	    {
/*printf("SCB wrong lun\n");*/
no_tag:
	    return tul_msgout_abort_targ(pHcs);
	    }
	pHcs->HCS_ActScb = pScb;
	if ((phase = tul_msgin_accept(pHcs)) == -1)
		return (-1);
	}
    return(tul_next_state(pHcs));
}



/***************************************************************************/
static long     tul_msgout_abort_targ(I91U_HCS *pHcs)
{
    int phase;

    TUL_WR(pHcs->HCS_Base + TUL_SSignal, ((TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7)) | TSC_SET_ATN));
    phase = tul_msgin_accept(pHcs);
    if (phase == -1)
	return(-1);
    if (phase != MSG_OUT)
	return(tul_bad_seq(pHcs));

    TUL_WR(pHcs->HCS_SFifo, MSG_ABORT);
    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);

    pHcs->HCS_Flags |= HCF_EXPECT_DISC;
    if (wait_tulip(pHcs) != -1) {
	return (tul_bad_seq(pHcs));
    }
    return -1;
}

/***************************************************************************/
static long     tul_msgout_abort_tag(I91U_HCS *pHcs)
{
    int phase;

    TUL_WR(pHcs->HCS_Base + TUL_SSignal, ((TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7)) | TSC_SET_ATN));
    phase = tul_msgin_accept(pHcs);
    if (phase == -1)
	return(-1);
    if (phase != MSG_OUT)
	return(tul_bad_seq(pHcs));

    TUL_WR(pHcs->HCS_SFifo, MSG_ABORT_TAG);
    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);

    pHcs->HCS_Flags |= HCF_EXPECT_DISC;
    if (wait_tulip(pHcs) != -1)
	{
	return (tul_bad_seq(pHcs));
	}

    return (-1);
}

/***************************************************************************/
static long     tul_msgin(I91U_HCS *pHcs)
{
    TCS         *pTcb = pHcs->HCS_ActTcs;
    long        cnt;
    int         phase;
    BYTE        imsg;

    for (;;)
	{
	if ((cnt = TUL_RD(pHcs->HCS_Base + TUL_SFifoCnt) & 0x1F) != 0)
	    {                           /* Flush SCSI FIFO              */
	    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
	    }
	TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, 1);
	TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_IN);
	if ((phase = wait_tulip(pHcs)) == -1)
	    return (-1);
	imsg = TUL_RD(pHcs->HCS_SFifo);
	switch (imsg) {
	    case MSG_DISC:               /* Disconnect msg */
		phase = tul_msgin_discon(pHcs);
		break;
	    case MSG_SDP:
	    case MSG_RESTORE:
	    case MSG_NOP:
		phase = tul_msgin_accept(pHcs);
		break;
	    case MSG_REJ:               /* Clear ATN first              */
		TUL_WR(pHcs->HCS_Base + TUL_SSignal, (TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7)));
		if ((pTcb->TCS_Flags & (TCF_SYNC_DONE | TCF_NO_SYNC_NEGO)) == 0)
		    {                   /* do sync nego */
		    TUL_WR(pHcs->HCS_Base + TUL_SSignal, ((TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7)) | TSC_SET_ATN));
		    }
		phase = tul_msgin_accept(pHcs);
		break;
	    case MSG_EXTEND:               /* extended msg */
		phase = tul_msgin_extend(pHcs);
		break;
	    case MSG_IGNOREWIDE:
		phase = tul_msgin_ignore_wid_resid(pHcs);
		break;
	    case MSG_COMP:
		pHcs->HCS_Flags |= HCF_EXPECT_DONE_DISC;
		TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
		if ((phase = tul_msgin_accept(pHcs)) != -1) {
		    return (tul_bad_seq(pHcs));
		}
		break;
	    default:
		phase = tul_msgout_reject(pHcs);
		break;
	}
	if (phase != MSG_IN)
	    return (phase);
    }
	/* statement won't reach here */
}


/***************************************************************************/
static long     tul_msgin_discon(I91U_HCS *pHcs)
{
    pHcs->HCS_Flags |= HCF_EXPECT_DISC;
    if (tul_msgin_accept(pHcs) == -1)
	return (-1);
    return (tul_bad_seq(pHcs));
}


/***************************************************************************/
static long     tul_msgout_reject(I91U_HCS *pHcs)
{
    int         phase;

    TUL_WR(pHcs->HCS_Base + TUL_SSignal, ((TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7)) | TSC_SET_ATN));

    if ((phase = tul_msgin_accept(pHcs)) == -1)
	return(-1);

    if (phase == MSG_OUT)
	{
	TUL_WR(pHcs->HCS_SFifo, MSG_REJ); /* Msg reject           */
	TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
	return (wait_tulip(pHcs));
	}
    return (phase);
}

/************** tul_msgout_sync rtn **************************/
static long  tul_msgout_sync(I91U_HCS *pHcs)
{ 
char    phase;
					 
    TCS         *pTcb = pHcs->HCS_ActTcs;

    pTcb->TCS_Flags |= TCF_SYNC_DONE;

    TUL_WR(pHcs->HCS_SFifo, MSG_EXTEND);
    TUL_WR(pHcs->HCS_SFifo, 3); /* extended msg length */
    TUL_WR(pHcs->HCS_SFifo, 1); /* sync request */
    TUL_WR(pHcs->HCS_SFifo, tul_rate_tbl[pTcb->TCS_Flags & TCF_SCSI_RATE]);
    TUL_WR(pHcs->HCS_SFifo, MAX_OFFSET); /* REQ/ACK offset */
    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
    phase = wait_tulip(pHcs);
    TUL_WR(pHcs->HCS_Base+TUL_SCtrl0,  TSC_FLUSH_FIFO);
    TUL_WR(pHcs->HCS_Base+TUL_SSignal, TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7));

    return (phase);
}

/***************************************************************************/
static long     tul_msgout_nop(I91U_HCS *pHcs)
{
    TUL_WR(pHcs->HCS_SFifo, MSG_NOP); /* msg nop */
    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
    return (wait_tulip(pHcs));
}

/***************************************************************************/
static long     tul_msgout_ide(I91U_HCS *pHcs)
{
    TUL_WR(pHcs->HCS_SFifo, MSG_IDE); /* Initiator Detected Error */
    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
    return (wait_tulip(pHcs));
}

/***************************************************************************/
static long tul_msgin_ignore_wid_resid(I91U_HCS *pHcs)
{
    int    phase;

    if ((phase = tul_msgin_accept(pHcs)) != MSG_IN)
	return (phase);

	/* get */
    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_IN);
    if (wait_tulip(pHcs) == -1)
	return -1;

    TUL_WR(pHcs->HCS_SFifo, 0);           /* put pad  */
    TUL_RD(pHcs->HCS_SFifo);       /* get IGNORE field */
    TUL_RD(pHcs->HCS_SFifo);       /* get pad */

    return (tul_msgin_accept(pHcs));

}

/***************************************************************************/
static long     tul_msgin_extend(I91U_HCS *pHcs)
{
    int         phase;
    BYTE        len, idx;
/*    BYTE      msg[10];*/

    if ((phase = tul_msgin_accept(pHcs)) != MSG_IN)
	return (phase);

					/* Get extended msg length      */
    TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, 1);
    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_IN);
    if (wait_tulip(pHcs) == -1)
	return (-1);
    len = TUL_RD(pHcs->HCS_SFifo);
    pHcs->HCS_Msg[0] = len;
    for (idx = 1 ; len != 0; len--)
	{
	if ((phase = tul_msgin_accept(pHcs)) != MSG_IN)
	    return (phase);
	TUL_WRLONG(pHcs->HCS_Base + TUL_SCnt0, 1);
	TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_IN);
	if (wait_tulip(pHcs) == -1)
	    return (-1);
	pHcs->HCS_Msg[idx++]  = TUL_RD(pHcs->HCS_SFifo);
	}
    if (pHcs->HCS_Msg[1] == 1)
	{                       /* if it's synchronous data transfer request */
	if (pHcs->HCS_Msg[0] != 3)        /* if length is not right */
	    return (tul_msgout_reject(pHcs));
	if (pHcs->HCS_ActTcs->TCS_Flags & TCF_NO_SYNC_NEGO)
	    {                   /* Set OFFSET=0 to do async, nego back */
	    pHcs->HCS_Msg[3] = 0;
	    }
	else
	    {
	    if ((tul_msgin_sync(pHcs) == 0) && 
		(pHcs->HCS_ActTcs->TCS_Flags & TCF_SYNC_DONE))
		{
		tul_sync_done(pHcs);
		return (tul_msgin_accept(pHcs));
		}
	    }

	TUL_WR(pHcs->HCS_Base + TUL_SSignal, ((TUL_RD(pHcs->HCS_Base + TUL_SSignal) & (TSC_SET_ACK | 7)) | TSC_SET_ATN));
	if ((phase  = tul_msgin_accept(pHcs)) != MSG_OUT)
	    return (phase);
					/* sync msg out */
	TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO);
/*      TUL_WR(pHcs->HCS_SFifo, TSC_FLUSH_FIFO);            */

	tul_sync_done(pHcs);

	TUL_WR(pHcs->HCS_SFifo, MSG_EXTEND);
	TUL_WR(pHcs->HCS_SFifo, 3);
	TUL_WR(pHcs->HCS_SFifo, 1);
	TUL_WR(pHcs->HCS_SFifo, pHcs->HCS_Msg[2]);
	TUL_WR(pHcs->HCS_SFifo, pHcs->HCS_Msg[3]);

	TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
	return (wait_tulip(pHcs));
	}

    if ((pHcs->HCS_Msg[0] != 2) || (pHcs->HCS_Msg[1] != 3))
	return  (tul_msgout_reject(pHcs));
					/* if it's WIDE DATA XFER REQ   */
    if (pHcs->HCS_ActTcs->TCS_Flags & TCF_NO_WDTR)
	{
	pHcs->HCS_Msg[2] = 0;
	}
    else
	{
	if (pHcs->HCS_Msg[2] > 2)    /* > 32 bits            */
	    return  (tul_msgout_reject(pHcs));
	if (pHcs->HCS_Msg[2] == 2)   /* == 32                */
	    {
	    pHcs->HCS_Msg[2] = 1;
	    }
	else
	    {
	    if ((pHcs->HCS_ActTcs->TCS_Flags & TCF_NO_WDTR) == 0)
		{
		tul_wdtr_done(pHcs);
		if ((pHcs->HCS_ActTcs->TCS_Flags & (TCF_SYNC_DONE | TCF_NO_SYNC_NEGO)) == 0)
		    TUL_WR(pHcs->HCS_Base + TUL_SSignal, ((TUL_RD(pHcs->HCS_Base +  TUL_SSignal) & (TSC_SET_ACK | 7)) | TSC_SET_ATN));
		return  (tul_msgin_accept(pHcs));
		}
	    }
	}
    TUL_WR(pHcs->HCS_Base + TUL_SSignal, ((TUL_RD(pHcs->HCS_Base +  TUL_SSignal) & (TSC_SET_ACK | 7)) | TSC_SET_ATN));

    if ((phase =  tul_msgin_accept(pHcs)) != MSG_OUT)
	return (phase);
					/* WDTR msg out                 */
    TUL_WR(pHcs->HCS_SFifo, MSG_EXTEND);
    TUL_WR(pHcs->HCS_SFifo, 2);
    TUL_WR(pHcs->HCS_SFifo, 3);
    TUL_WR(pHcs->HCS_SFifo, pHcs->HCS_Msg[2]);
    TUL_WR(pHcs->HCS_SCmd, TSC_XF_FIFO_OUT);
    return (wait_tulip(pHcs));
}

/***************************************************************************/
static long     tul_msgin_sync(I91U_HCS *pHcs)
{
    char default_period;

    default_period = tul_rate_tbl[pHcs->HCS_ActTcs->TCS_Flags & TCF_SCSI_RATE];
    if (pHcs->HCS_Msg[3] > MAX_OFFSET)
	{
	pHcs->HCS_Msg[3] = MAX_OFFSET;
	if (pHcs->HCS_Msg[2] < default_period)
	    {
	    pHcs->HCS_Msg[2] = default_period;
	    return 1;
	    }                                                                                    
	if (pHcs->HCS_Msg[2] >= 59)
	    {                           /* Change to async              */
	    pHcs->HCS_Msg[3] = 0;
	    }
	return 1;
	}

		/* offset requests asynchronous transfers ? */
    if (pHcs->HCS_Msg[3] == 0)
	{
	return 0;
	}
    if (pHcs->HCS_Msg[2] < default_period)
	{
	pHcs->HCS_Msg[2] = default_period;
	return 1;
	}

    if (pHcs->HCS_Msg[2] >= 59)
	{
	pHcs->HCS_Msg[3] = 0;
	return 1;
	}
    return 0;
}


/***************************************************************************/
static void  tul_wdtr_done(I91U_HCS *pHcs)            
{
    pHcs->HCS_ActTcs->TCS_Flags &= ~TCF_SYNC_DONE;
    pHcs->HCS_ActTcs->TCS_Flags |= TCF_WDTR_DONE;

    pHcs->HCS_ActTcs->TCS_JS_Period = 0;
    if (pHcs->HCS_Msg[2])
	{                               /* if 16 bit */
	pHcs->HCS_ActTcs->TCS_JS_Period |= TSC_WIDE_SCSI;
	}
    pHcs->HCS_ActTcs->TCS_SConfig0 &= ~TSC_ALT_PERIOD;
    TUL_WR(pHcs->HCS_Base+TUL_SConfig, pHcs->HCS_ActTcs->TCS_SConfig0);
    TUL_WR(pHcs->HCS_Base+TUL_SPeriod, pHcs->HCS_ActTcs->TCS_JS_Period);

}

/***************************************************************************/
void     tul_sync_done(I91U_HCS *pHcs)
{
    int i;

    pHcs->HCS_ActTcs->TCS_Flags |= TCF_SYNC_DONE;

    if (pHcs->HCS_Msg[3])
	{
	pHcs->HCS_ActTcs->TCS_JS_Period |= pHcs->HCS_Msg[3];
	for (i = 0; i < 8; i++)
	    {
	    if (tul_rate_tbl[i] >= pHcs->HCS_Msg[2])  /* pick the big one */
		break;
	    }
	pHcs->HCS_ActTcs->TCS_JS_Period |= (i << 4);
	pHcs->HCS_ActTcs->TCS_SConfig0 |= TSC_ALT_PERIOD;
	}
    TUL_WR(pHcs->HCS_Base+TUL_SConfig, pHcs->HCS_ActTcs->TCS_SConfig0);
    TUL_WR(pHcs->HCS_Base+TUL_SPeriod, pHcs->HCS_ActTcs->TCS_JS_Period);

}


/***************************************************************************/
static long     tul_reset_scsi(I91U_HCS *pHcs)
{
    ULONG       i;

    TUL_WR( pHcs->HCS_Base + TUL_SCtrl0, TSC_RST_BUS);

    while(!((pHcs->HCS_JSInt = TUL_RD(pHcs->HCS_Base + TUL_SInt)) & TSS_SCSIRST_INT));
	/* reset tulip chip */

    TUL_WR(pHcs->HCS_Base + TUL_SSignal, 0);

	/* Stall for a while, wait for target's firmware ready,make it 2 sec ! */
	/* SONY 5200 tape drive won't work if only stall for 1 sec */
    for (i = 0; i < 1000; i++)         /* wait 1 sec                   */
	DELAY (1000);  /* delay 1 msec */

    TUL_RD(pHcs->HCS_Base + TUL_SInt);

    for (i = 0; i < 16000; i++)         
	DELAY (1000);  /* delay 1 msec */

    return (1);
}

static long     tul_post_scsi_rst(I91U_HCS *pHcs)
{
    I91U_SCB *pScb;
    TCS *pTcb;
    int i;

    pHcs->HCS_ActScb = 0;

    pScb = pHcs->HCS_Scb ;
    for (i = 0; i < pHcs->HCS_NumScbs; i++, pScb++)
	{
	if (pScb->SCB_Status & SCB_BUSY) 
	    {
	    tul_unlink_busy_scb(pHcs, pScb);

	    pScb->SCB_HaStat = HOST_SCSI_RST;
	    tul_append_done_scb(pHcs, pScb);
	    } 
	}

					/* clear sync done flag         */
    pTcb = &pHcs->HCS_Tcs[0];
    for (i = 0; i < pHcs->HCS_MaxTar; pTcb++,  i++)
	{
	pTcb->TCS_Flags &= ~(TCF_SYNC_DONE | TCF_WDTR_DONE);
		/* Initialize the sync. xfer register values to an asyn xfer */
	pTcb->TCS_JS_Period = 0;
	pTcb->TCS_SConfig0 = pHcs->HCS_SConf1;
	pTcb->TCS_TagCnt = 0;
	pTcb->TCS_NonTagScb = NULL;
	} /* for */
    return (-1);
}

/***************************************************************************/
static void    tul_select_atn_stop(I91U_HCS *pHcs, I91U_SCB *pScb)
{

    TUL_WR(pHcs->HCS_SCmd, TSC_SELATNSTOP);
    pScb->SCB_Status |= SCB_SELECT;
    pScb->SCB_NxtStat = 1;
    pHcs->HCS_ActScb = pScb;
}


/***************************************************************************/
static void    tul_select_atn(I91U_HCS *pHcs, I91U_SCB *pScb)
{
#if 0
    int i;


    TUL_WR(pHcs->HCS_SFifo, pScb->SCB_Ident);
    for (i = 0; i < (int)pScb->SCB_CDBLen; i++)
	TUL_WR(pHcs->HCS_SFifo, pScb->SCB_CDB[i]);
#else
    TUL_WR(pHcs->HCS_SFifo, pScb->SCB_Ident);
    outsb(pHcs->HCS_SFifo, pScb->SCB_CDB, pScb->SCB_CDBLen);
#endif
    TUL_WR(pHcs->HCS_SCmd, TSC_SEL_ATN);

    pScb->SCB_Status |= SCB_SELECT;
    pScb->SCB_NxtStat = 2;
    pHcs->HCS_ActScb = pScb;
}

/***************************************************************************/
static void    tul_select_atn3(I91U_HCS *pHcs, I91U_SCB *pScb)
{
#if 0
    int         i;

    TUL_WR(pHcs->HCS_SFifo, pScb->SCB_Ident);
    TUL_WR(pHcs->HCS_SFifo, pScb->SCB_TagMsg);
    TUL_WR(pHcs->HCS_SFifo, pScb->SCB_TagId);
    for (i = 0; i < (int) pScb->SCB_CDBLen; i++)
	TUL_WR(pHcs->HCS_SFifo, pScb->SCB_CDB[i]);
#else
    outsb(pHcs->HCS_SFifo, &(pScb->SCB_Ident), pScb->SCB_CDBLen+3);
#endif
    TUL_WR(pHcs->HCS_SCmd, TSC_SEL_ATN3);

    pScb->SCB_Status |= SCB_SELECT;
    pScb->SCB_NxtStat = 2;
    pHcs->HCS_ActScb = pScb;

}


/***************************************************************************/
static long     tul_msgin_accept(I91U_HCS *pHcs)
{
    TUL_WR(pHcs->HCS_SCmd, TSC_MSG_ACCEPT);
    return (wait_tulip(pHcs));
}

/***************************************************************************/
static long     wait_tulip(I91U_HCS *pHcs)
{
    I91U_SCB         *pScb = pHcs->HCS_ActScb;
/*    register BYTE s;
    int phase;*/

    while (!((pHcs->HCS_JSStatus0 = TUL_RD(pHcs->HCS_Base + TUL_SStatus0))
			 & TSS_INT_PENDING));

    pHcs->HCS_Phase = pHcs->HCS_JSStatus0 & TSS_PH_MASK;
    pHcs->HCS_JSInt = TUL_RD(pHcs->HCS_Base + TUL_SInt);
    pHcs->HCS_JSStatus1 = TUL_RD(pHcs->HCS_Base + TUL_SStatus1);

/* printf("int=%x ps=%x ", pHcs->HCS_JSInt, pHcs->HCS_Phase);	*/

    if (pHcs->HCS_JSInt & TSS_SCSIRST_INT)
	{                               /* if SCSI bus reset detected   */
	return (int_tul_scsi_rst(pHcs));
	}

    if (pHcs->HCS_JSInt & TSS_RESEL_INT)
	{                               /* if SCSI bus reset detected   */
	return (int_tul_resel(pHcs));
	}

    if (pHcs->HCS_JSInt & TSS_SEL_TIMEOUT)
	{                         /* if selected/reselected timeout interrupt */
	return (int_tul_busfree(pHcs));
	}

    if (pHcs->HCS_JSInt & TSS_DISC_INT) {     /* BUS disconnection            */
	if (pHcs->HCS_Flags & HCF_EXPECT_DONE_DISC) {
	    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO); /* Flush SCSI FIFO  */

	    TUL_WR(pHcs->HCS_Base + TUL_SConfig, TSC_INITDEFAULT);
	    TUL_WR(pHcs->HCS_Base + TUL_SCtrl1, TSC_HW_RESELECT);/* Enable HW reselect       */

	    tul_unlink_busy_scb(pHcs, pScb);
	    tul_append_done_scb(pHcs, pScb);
	    pHcs->HCS_ActScb = NULL;
	    pHcs->HCS_Flags &= ~HCF_EXPECT_DONE_DISC;

	    return (-1);
	    }

	if (pHcs->HCS_Flags & HCF_EXPECT_DISC) {
	    TUL_WR(pHcs->HCS_Base + TUL_SCtrl0, TSC_FLUSH_FIFO); /* Flush SCSI FIFO  */
	    TUL_WR(pHcs->HCS_Base + TUL_SConfig, TSC_INITDEFAULT);
	    TUL_WR(pHcs->HCS_Base + TUL_SCtrl1, TSC_HW_RESELECT);/* Enable HW reselect       */
	    pHcs->HCS_ActScb = NULL;
	    pHcs->HCS_Flags &= ~HCF_EXPECT_DISC;
	    return (-1);
	    }
	return (int_tul_busfree(pHcs));
	}

	return (pHcs->HCS_Phase);

}

/**************************** EOF *********************************/
