/**************************************************************************
**
**  $Id: i91u.c,v 1.82.2.9 1998/01/08 12:04:00 se Exp $
**
**  Device driver for the   INI-9XXXU/UW or INIC-940/950  PCI SCSI Controller.
**
**  FreeBSD  
**
**-------------------------------------------------------------------------
**
**  Written for 386bsd and FreeBSD by
**	Winston Hung		<winstonh@initio.com>
**
**
**-------------------------------------------------------------------------
**
** Copyright (c) 1997-99 Initio Corp.  All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 3. The name of the author may not be used to endorse or promote products
**    derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
** IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
** IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
** INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
** NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
** THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
***************************************************************************
*/



#include <sys/param.h>

#ifdef KERNEL
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/buf.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <vm/vm.h>
#include <vm/pmap.h>
#include <vm/vm_extern.h>
#endif /* KERNEL */

#include <pci/pcivar.h>
#include <pci/pcireg.h>

#include <cam/cam.h>
#include <cam/cam_conf.h>
#include <cam/cam_ccb.h>
#include <cam/cam_sim.h>
#include <cam/cam_xpt.h>
#include <cam/cam_xpt_sim.h>
#include <cam/cam_xpt_periph.h>
#include <cam/cam_debug.h>

#include <cam/scsi/scsi_all.h>
#include <cam/scsi/scsi_message.h>

#include <machine/bus_pio.h>
#include <machine/bus.h>
#include <machine/clock.h>


/*  #include <pci/ini9100u.h>  */

#define INT32	  int32
#define U_INT32   u_int32

#ifdef __alpha__
/* XXX */
#undef vtophys
#define	vtophys(va)	(pmap_kextract(((vm_offset_t) (va))) \
			 + 1*1024*1024*1024)
#endif

#define OutB(val, port) 		outb(port, val)
#define OutW(val, port) 		outw(port, val)
#define OutL(val, port) 		outl(port, val)

#define PCI_DEVICE_ID_INIC940 	0x94001101ul
#define PCI_DEVICE_ID_INIC941 	0x94011101ul
#define PCI_DEVICE_ID_INIC950 	0x95001101ul
#define PCI_BASEADDR0			0x10


#ifdef KERNEL

/* static void i91u_Interrupt (I91U_HCS *pHcs);  */


static	char*	i91u_probe( pcici_t tag, pcidi_t type);
static	void	i91u_attach( pcici_t tag, int unit);

static void i91u_timeout( void *arg1);



static  void i91u_action(struct cam_sim *sim, union ccb *ccb);
static  void i91u_poll(struct cam_sim *sim);



#endif /* KERNEL */




static u_long	i91u_count;

struct	pci_device   i91u_device = {
	"iha",
	i91u_probe,
	i91u_attach,
	&i91u_count,
	NULL
};

DATA_SET (pcidevice_set, i91u_device);







#include <pci/i91uscsi.c>





/*
 * We have a scb which has been processed by the
 * adaptor, now we look to see how the operation
 * went.
 */
static void
i91u_done(I91U_HCS *pHcs, I91U_SCB *pScb)
{
	union ccb *ccb = (union ccb *)pScb->SCB_ccb;
	struct ccb_scsiio *csio;

	pScb->SCB_ccb = NULL;
#if 0
printf("done:hs=%x ts=%x\n",pScb->SCB_HaStat ,pScb->SCB_TaStat );
tul_do_pause(1000);
#endif

	if((ULONG)ccb == 0) {
		printf("free scb while ccb=0 in i91u_done \n");
		tul_release_scb(pHcs, pScb);
		return;
	}

	untimeout(i91u_timeout, (caddr_t) pScb,ccb->ccb_h.timeout_ch);
#if 0
	if ((ccb->ccb_h.flags & CAM_DIR_MASK) != CAM_DIR_NONE) {
		bus_dmasync_op_t op;

		if ((ccb->ccb_h.flags & CAM_DIR_MASK) == CAM_DIR_IN)
			op = BUS_DMASYNC_POSTREAD;
		else
			op = BUS_DMASYNC_POSTWRITE;
		bus_dmamap_sync(pHcs->dmat, pScb->SCB_dmamap, op);
		bus_dmamap_unload(pHcs->buffer_dmat, pScb->SCB_dmamap);
	}
#endif

	csio = &ccb->csio;
	/*
	**	Check the status.
	*/
	if (   (pScb->SCB_HaStat == HOST_OK)
		&& (pScb->SCB_TaStat == TARGET_GOOD)) {

		ccb->ccb_h.status = CAM_REQ_CMP;
		csio->scsi_status = SCSI_STATUS_OK;

	} else if (   (pScb->SCB_HaStat == HOST_OK)
		&& (pScb->SCB_TaStat== TARGET_CHKCOND  )) {

		/*
		**   Check condition code
		*/
		ccb->ccb_h.status  |= CAM_AUTOSNS_VALID;
		csio->scsi_status = SCSI_STATUS_CHECK_COND;

	} else if ((pScb->SCB_HaStat== HOST_OK)
		&& (pScb->SCB_TaStat== TARGET_BUSY)) {
		/*
		**   Target is busy.
		*/
		ccb->ccb_h.status = CAM_SCSI_STATUS_ERROR;
		csio->scsi_status = SCSI_STATUS_BUSY;

	} else if ((pScb->SCB_HaStat== HOST_OK)
		&& (pScb->SCB_TaStat== TARGET_QFULL)) {

		/*
		**   Target is queue full.
		*/
		ccb->ccb_h.status = CAM_SCSI_STATUS_ERROR;
		csio->scsi_status = SCSI_STATUS_QUEUE_FULL;


	} else if (pScb->SCB_HaStat== HOST_SEL_TOUT   ) {

		/*
		**   Device failed selection
		*/
		ccb->ccb_h.status = CAM_SEL_TIMEOUT;

	} else { /*  Other protocol messes */
		ccb->ccb_h.status = CAM_CMD_TIMEOUT;
	}


	tul_release_scb(pHcs, pScb);
	xpt_done(ccb);
}

static void
i91u_timeout(void *arg1)
{

	I91U_SCB *pScb = (I91U_SCB*) arg1;
	I91U_HCS *pHcs;
	union ccb *ccb;
	int     unit;
	I91U_HCS *aic;
	int s;


	ccb = (union ccb *)pScb->SCB_ccb;
	if (ccb == NULL) 
		return;

	pHcs = (I91U_HCS *)xpt_path_sim(ccb->ccb_h.path)->softc;


	s = splcam();
	if (tul_abort_ccb(pHcs, (ULONG) ccb)) {
		printf("iha: timed out\n");

		ccb->ccb_h.status |= CAM_REQ_ABORTED;
		xpt_done(ccb);
	}
	splx(s);
}

static void i91u_exec_scb(void *arg, bus_dma_segment_t *dm_segs, int nseg, int error)
{
	ULONG	bScbIdx;
	I91U_SCB * pScb;
	union ccb *ccb;
	I91U_HCS *pHcs;
	I91U_SG *sg;
	struct ccb_scsiio *csio;
	int s;

	pScb = (I91U_SCB *)arg;
	ccb = (union ccb *)pScb->SCB_ccb;
	csio = &ccb->csio;
	pHcs = (I91U_HCS *)xpt_path_sim(ccb->ccb_h.path)->softc;

	if(pScb->SCB_CDB[0] == 0x3) {

		/* disable SCSI DISCONNECT if it is "REQUEST SENSE" command */
		printf("--REQUEST SENSE--");
     		pScb->SCB_Ident &= 0xBF;

	}


	if (error != 0) {
		if (error != EFBIG)
			printf("iha: Unexepected error 0x%x returned from bus_dmamap_load\n", error);
		if (ccb->ccb_h.status == CAM_REQ_INPROG) {
			xpt_freeze_devq(ccb->ccb_h.path, /*count*/1);
			ccb->ccb_h.status = CAM_REQ_TOO_BIG|CAM_DEV_QFRZN;
		}
		tul_release_scb(pHcs, pScb);
		xpt_done(ccb);
		return;
	}





/***	printf("--exec(seg=%d,xferl=%d,cdl=%d,tag=%02x)--",nseg,csio->dxfer_len,pScb->SCB_CDBLen,pScb->SCB_TagMsg);***/	

	pScb->SCB_BufLen = csio->dxfer_len;

	if (nseg == 1) {
		pScb->SCB_BufPtr = (ULONG)dm_segs->ds_addr;
	}
	else if (nseg != 0) {
		bus_dma_segment_t *end_seg;
		bus_dmasync_op_t op;

		sg = pScb->SCB_SGList;
		end_seg = dm_segs + nseg;

		while (dm_segs < end_seg) {
			sg->SG_Len = dm_segs->ds_len;
			sg->SG_Ptr = dm_segs->ds_addr;
			sg++;
			dm_segs++;
		}

		pScb->SCB_Flags	|= SCF_SG;
		pScb->SCB_BufPtr = (ULONG)pScb->SCB_SGPAddr;
		pScb->SCB_SGLen = nseg;
		
#if 0
		if ((ccb->ccb_h.flags & CAM_DIR_MASK) == CAM_DIR_IN)
			op = BUS_DMASYNC_PREREAD;
		else
			op = BUS_DMASYNC_PREWRITE;

		bus_dmamap_sync(pHcs->buffer_dmat, pScb->SCB_dmamap, op);
#endif

	}

	s = splcam();
	ccb->ccb_h.timeout_ch = timeout(i91u_timeout, (caddr_t)pScb, (ccb->ccb_h.timeout * hz)/1000);

	tul_exec_scb(pHcs, pScb);
	splx(s);
}

/*==========================================================
**
**
**	Start execution of a SCSI command.
**	This is called from the generic SCSI driver.
**
**
**==========================================================
*/
void i91u_action(struct cam_sim *sim, union ccb *ccb)
{
    I91U_SCB  *pScb;
    I91U_HCS  *pHcs;
    int	  flags;
/*    struct scsi_inquiry_data *inq_buf;	*/

    pHcs = (I91U_HCS  *) cam_sim_softc(sim);

    switch (ccb->ccb_h.func_code) {
	/* Common cases first */
	case XPT_SCSI_IO:	/* Execute the requested I/O operation */
	case XPT_RESET_DEV:	/* Bus Device Reset the specified SCSI device */
	{
	    struct ccb_scsiio *csio = &ccb->csio;

	    flags = ccb->ccb_h.flags;

	    /*
	     * Last time we need to check if this CCB needs to
	     * be aborted.
	     */
	    if ((ccb->ccb_h.status & CAM_STATUS_MASK) != CAM_REQ_INPROG) {
			xpt_done(ccb);
			return;
	    }
	    ccb->ccb_h.status |= CAM_SIM_QUEUED;

	    /*
	     * get an scb to use. If the transfer
	     * is from a buf (possibly from interrupt time)
	     * then we can't allow it to sleep
	     */
	    if ((pScb = tul_alloc_scb(pHcs)) == NULL) {
/***			printf("--!pScb--");***/
			ccb->ccb_h.status = CAM_RESRC_UNAVAIL;
			xpt_done(ccb);
			return;
	    }

	    pScb->SCB_ccb =(ULONG)ccb;

	    pScb->SCB_Target = ccb->ccb_h.target_id;
	    pScb->SCB_Lun    = ccb->ccb_h.target_lun;
            pScb->SCB_Flags = SCF_POST;
			/* After SCSI done, call post routine*/
	    pScb->SCB_Post = (void *)i91u_done;

	    *((UWORD *)&pScb->SCB_HaStat) = 0;

	    if (ccb->ccb_h.func_code == XPT_RESET_DEV) {
		    pScb->SCB_Opcode  = BusDevRst;
		    tul_exec_scb(pHcs, pScb);
	    } 
	    else {
	    	pScb->SCB_Opcode  = ExecSCSI;

		if((flags & CAM_DIR_MASK) == CAM_DIR_NONE)
			pScb->SCB_Flags |= SCF_NO_XF;
		else if(flags & CAM_DIR_IN)
			pScb->SCB_Flags |= SCF_DIN;
		else if(flags & CAM_DIR_OUT)
			pScb->SCB_Flags |= SCF_DOUT;

		pScb->SCB_Ident     = pScb->SCB_Lun | DISC_ALLOW;
		pScb->SCB_CDBLen = csio->cdb_len;
		if (pScb->SCB_CDBLen > 12) {
			ccb->ccb_h.status = CAM_REQ_INVALID;
			tul_release_scb(pHcs, pScb);
			xpt_done(ccb);
		}

		if ((flags & CAM_CDB_POINTER) != 0) {
			if ((flags & CAM_CDB_PHYS) == 0) {
				bcopy(csio->cdb_io.cdb_ptr, pScb->SCB_CDB, pScb->SCB_CDBLen);
			} else {
				ccb->ccb_h.status = CAM_REQ_INVALID;
				tul_release_scb(pHcs, pScb);
				xpt_done(ccb);
				return;
			}
		} else {
			bcopy(csio->cdb_io.cdb_bytes, pScb->SCB_CDB, pScb->SCB_CDBLen);
		}

	        if (!(ccb->ccb_h.flags & CAM_DIS_AUTOSENSE)) {
		    pScb->SCB_Flags |= SCF_SENSE;
		    pScb->SCB_SenseLen = csio->sense_len;
		    pScb->SCB_SensePtr = vtophys(&csio->sense_data);
		}

		if((flags & CAM_TAG_ACTION_VALID) != 0 && (csio->tag_action != CAM_TAG_ACTION_NONE)){
		    	pScb->SCB_TagMsg = csio->tag_action;
		}
		else{
		    	pScb->SCB_TagMsg = 0;
		}



	        /* Only use S/G if there is a transfer */
		if ((flags & CAM_DIR_MASK) != CAM_DIR_NONE) {
	
		    if ((flags & CAM_SCATTER_VALID) == 0) {
			/*
			 * We've been given a pointer
			 * to a single buffer.
			 */
			if ((flags & CAM_DATA_PHYS)==0) {
				int s;
				int error;

/***				printf("--CAM_DATA_PHYS=0--");***/
				s = splsoftvm();
				error = bus_dmamap_load(
				    pHcs->dmat,
#if 0
				    pScb->SCB_dmamap,
#else
				    pHcs->dmamap,
#endif
				    csio->data_ptr,
				    csio->dxfer_len,
				    i91u_exec_scb,
				    pScb,
				    /*flags*/0);
				if (error == EINPROGRESS) {
					/*
					 * So as to maintain
					 * ordering, freeze the
					 * controller queue
					 * until our mapping is
					 * returned.
					 */
					xpt_freeze_simq(pHcs->sim,
						1);
					csio->ccb_h.status |=
					    CAM_RELEASE_SIMQ;
				}
				splx(s);
			} else {
				struct bus_dma_segment seg; 

				/* Pointer to physical buffer */
				seg.ds_addr =
				    (bus_addr_t)csio->data_ptr;
				seg.ds_len = csio->dxfer_len;
				i91u_exec_scb(pScb,&seg,1,0);
			}
		    } else {
				struct bus_dma_segment *segs;

				if ((flags & CAM_DATA_PHYS) != 0)
					panic("i91u_action - Physical "
					      "segment pointers "
					      "unsupported");

				if ((flags&CAM_SG_LIST_PHYS)==0)
					panic("i91u_action - Virtual "
					      "segment addresses "
					      "unsupported");

				segs = (struct bus_dma_segment *)
					    csio->data_ptr;
				i91u_exec_scb(pScb,segs,csio->sglist_cnt,0);

			}
		} else {
			int s;

			s = splcam();
			ccb->ccb_h.timeout_ch = timeout(i91u_timeout, (caddr_t)pScb, (ccb->ccb_h.timeout * hz)/1000);

			tul_exec_scb(pHcs, pScb);
			splx(s);
		}	
	    }
	    break;
	}

	case XPT_EN_LUN:		/* Enable LUN as a target */
	case XPT_TARGET_IO:		/* Execute target I/O request */
	case XPT_ACCEPT_TARGET_IO:	/* Accept Host Target Mode CDB */
	case XPT_CONT_TARGET_IO:	/* Continue Host Target I/O Connection*/
	case XPT_ABORT:			/* Abort the specified CCB */
		/* XXX Implement */
		ccb->ccb_h.status = CAM_REQ_INVALID;
		xpt_done(ccb);
		break;

	case XPT_SET_TRAN_SETTINGS:
	{
		/* XXX Implement */
		ccb->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;
	}

	case XPT_GET_TRAN_SETTINGS:
	/* Get default/user set transfer settings for the target */
	{

		struct	ccb_trans_settings *cts;

		cts = &ccb->cts;

		cts->flags = 0;
		cts->flags |= CCB_TRANS_DISC_ENB;

		cts->flags |= CCB_TRANS_TAG_ENB;

		cts->bus_width = (pHcs->HCS_MaxTar == 16) ?
			MSG_EXT_WDTR_BUS_16_BIT : MSG_EXT_WDTR_BUS_8_BIT;

		cts->sync_period = 12;
		cts->sync_offset = 15;

		cts->valid = CCB_TRANS_SYNC_RATE_VALID
			   | CCB_TRANS_SYNC_OFFSET_VALID
			   | CCB_TRANS_BUS_WIDTH_VALID
			   | CCB_TRANS_DISC_VALID
			   | CCB_TRANS_TQ_VALID;

		ccb->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;
	}

	case XPT_CALC_GEOMETRY:
	{
		struct	  ccb_calc_geometry *ccg;
		u_int32_t size_mb;
		u_int32_t secs_per_cylinder;

		ccg = &ccb->ccg;
		size_mb = ccg->volume_size
			/ ((1024L * 1024L) / ccg->block_size);
		
		if (size_mb >= 1024 ) {
			ccg->heads = 255;
			ccg->secs_per_track = 63;
		} else {
			ccg->heads = 64;
			ccg->secs_per_track = 32;
		}
		secs_per_cylinder = ccg->heads * ccg->secs_per_track;
		ccg->cylinders = ccg->volume_size / secs_per_cylinder;
		ccb->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;
	}

	case XPT_RESET_BUS:		/* Reset the specified SCSI bus */
	{
		tul_reset_scsi(pHcs);	/* Do Hardware Reset &	*/
		ccb->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;
	}
	case XPT_TERM_IO:		/* Terminate the I/O process */
		/* XXX Implement */
		ccb->ccb_h.status = CAM_REQ_INVALID;
		xpt_done(ccb);
		break;
	case XPT_PATH_INQ:		/* Path routing inquiry */
	{
		struct ccb_pathinq *cpi = &ccb->cpi;
		
		cpi->version_num = 1; /* XXX??? */
		cpi->hba_inquiry = PI_SDTR_ABLE;
		if (ccb->csio.tag_action != 0)
			cpi->hba_inquiry |= PI_TAG_ABLE;
		if(pHcs->HCS_MaxTar == 16)
			cpi->hba_inquiry |= PI_WIDE_16;
			
		cpi->target_sprt = 0;
		cpi->hba_misc = 0;
		cpi->hba_eng_cnt = 0;
		cpi->max_target = pHcs->HCS_MaxTar-1;
		cpi->max_lun = 7;
		cpi->initiator_id = pHcs->HCS_SCSI_ID;
		cpi->bus_id = cam_sim_bus(sim);
		strncpy(cpi->sim_vid, "FreeBSD", SIM_IDLEN);
		strncpy(cpi->hba_vid, "Initio", HBA_IDLEN);
		strncpy(cpi->dev_name, cam_sim_name(sim), DEV_IDLEN);
		cpi->unit_number = cam_sim_unit(sim);
		cpi->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;
	}
	default:
		ccb->ccb_h.status = CAM_REQ_INVALID;
		xpt_done(ccb);
		break;

	}
}

void i91u_poll(struct cam_sim *sim)
{
	I91U_HCS  *pHcs;

	pHcs = (I91U_HCS  *) cam_sim_softc(sim);

	tul_isr(pHcs);
}


void
i91u_attach (pcici_t config_id, int unit)
{

    u_int16_t io_port;
    u_int32_t id;

    UCHAR   irq;
    ULONG   wlval;
    I91U_HCS    *pHcs = 0;
    I91U_SCB    *pScb;
    int     error;
    int     s;
    int     openings;
    struct cam_devq *devq;
    union ccb	*work_ccb;

    if( unit >= MAX_SUPPORTED_ADAPTERS)
	return;

#if 0
    if( pHCS[unit] )
	return;
#endif

    io_port = 0;
    if (pci_map_port(config_id, PCI_BASEADDR0, &io_port) == 0)
	return;


    pHcs = (I91U_HCS *) malloc (sizeof (I91U_HCS), M_DEVBUF, M_NOWAIT);
    if( !pHcs ) {
	    printf("iha%d: cannot allocate HCS !\n", unit);
	    return;
    }

    bzero (pHcs, sizeof (I91U_HCS));
    pHcs->HCS_Base = io_port;
    pHcs->HCS_Unit = unit;

    if (!(pci_map_int(config_id, (void *)tul_isr, (void *)pHcs, &bio_imask))) {
		free(pHcs, M_DEVBUF);
		return;
    }



    switch ((id = pci_conf_read(config_id, PCI_ID_REG))) {

	    case PCI_DEVICE_ID_INIC940:
	    case PCI_DEVICE_ID_INIC941:
	    case PCI_DEVICE_ID_INIC950:
		break;

	    default:
		free(pHcs, M_DEVBUF);
		return;
    }

    /* Allocate a dmatag for our transfer DMA maps */
    /* DMA tag for mapping buffers into device visible space. */
    /* XXX Should be a child of the PCI bus dma tag */
    error = bus_dma_tag_create(/*parent*/ NULL,
				/*alignment*/ 0,
				/*boundary*/ 0,
				/*lowaddr*/ (0xFFFFFFFFL),
				/*highaddr*/ BUS_SPACE_MAXADDR,
				/*filter*/ NULL,
				/*filterarg*/ NULL,
				/*maxsize*/ MAXBSIZE,
				/*nsegments*/ I91U_SG_ENTRY		,
				/*maxsegsz*/ BUS_SPACE_MAXSIZE_24BIT,
				/*flags*/ BUS_DMA_ALLOCNOW,
				   &pHcs->dmat);
 
    if (error != 0) {
	printf("iha: Could not allocate DMA tag - error %d\n",
	       error);
	free(pHcs, M_DEVBUF);
	return;
    }

    pScb = (I91U_SCB *) malloc (sizeof (I91U_SCB)*MAX_SCB, M_DEVBUF, M_NOWAIT);
    if( !pScb ) {
	    printf("iha%d: cannot allocate SCB !\n", unit);
	    bus_dma_tag_destroy(pHcs->dmat);
	    free(pHcs, M_DEVBUF);
	    return;
    }
    bzero (pScb, sizeof (I91U_HCS)*MAX_SCB);

    init_tulip(pHcs, pScb, MAX_SCB);

    if (!(pci_map_int(config_id, (void *)tul_isr, (void *)pHcs, &cam_imask))) {
	printf("iha%d: Can't pci_map_int \n", unit);
	bus_dma_tag_destroy(pHcs->dmat);
	free(pScb, M_DEVBUF);
	free(pHcs, M_DEVBUF);
	return;
    }


    /*
    ** Create the device queue.  We only allow MAX_START-1 concurrent
    ** transactions so we can be sure to have one element free in our
    ** start queue to reset to the idle loop.
    */
    s = splcam();
    openings = MAX_SCB-1;
    devq = cam_simq_alloc(openings);
    if (devq == NULL){
	splx(s);
	return;
    }

    /*
    **	Now tell the generic SCSI layer
    **	about our bus.
    */
    pHcs->sim = cam_sim_alloc(i91u_action, i91u_poll, "iha", pHcs,
				pHcs->HCS_Unit ,
				1, openings, devq);
    if (pHcs->sim == NULL) {
	cam_simq_free(devq);
	splx(s);
	return;
    }

    if (xpt_bus_register(pHcs->sim, 0) != CAM_SUCCESS) {
	cam_sim_free(pHcs->sim, TRUE);
	splx(s);
	return;
    }

#ifdef __alpha__
    alpha_register_pci_scsi(config_id->bus, config_id->slot, pHcs->sim);
#endif	

    work_ccb = xpt_alloc_ccb();
    if (xpt_create_path(&pHcs->path, NULL,
		    cam_sim_path(pHcs->sim), CAM_TARGET_WILDCARD,
		    CAM_LUN_WILDCARD) != CAM_REQ_CMP) {
	xpt_bus_deregister(cam_sim_path(pHcs->sim));
	cam_sim_free(pHcs->sim, TRUE);
	xpt_free_ccb(work_ccb);
	splx(s);
	return;
    }

    splx(s);
    return;

}





static	char*
i91u_probe (pcici_t tag, pcidi_t type)
{
	switch (type) {
	case PCI_DEVICE_ID_INIC940:
		return ("Initio INIC-941 PCI SCSI adapter");
	case PCI_DEVICE_ID_INIC941:
		return ("Initio INIC-940 PCI SCSI adapter");
	case PCI_DEVICE_ID_INIC950:
		return ("Initio INI-9XXXU/UW PCI SCSI adapter");
	}
	return (NULL);
}




